import asyncio
from datetime import datetime
import json
import re
import time
from typing import List

import arrow
import click
from rich.table import Table
from rich.text import Text

from grid.cli import rich_click

# Not sure why we have this import Console
# When I import it normally from rich I have
# import issues
from grid.cli.cli.utilities import create_openapi_object, string2dict
from grid.cli.core.callbacks import arrow_time_callback
from grid.cli.core.formatter import Formatable, print_to_console
from grid.openapi import (
    Externalv1Cluster,
    V1AWSClusterDriverSpec,
    V1ClusterDriver,
    V1ClusterPerformanceProfile,
    V1ClusterSpec,
    V1ClusterState,
    V1ClusterType,
    V1CreateClusterRequest,
    V1InstanceSpec,
    V1KubernetesClusterDriver,
)
from grid.openapi.api import cluster_service_api
from grid.sdk.client import create_swagger_client, create_websocket_client
from grid.websocket.model import ClusterLogsRequest
from grid.websocket.service import ClusterService

CLUSTER_STATE_CHECKING_TIMEOUT = 60
MAX_CLUSTER_WAIT_TIME = 5400


class ClusterList(Formatable):
    def __init__(self, clusters: [Externalv1Cluster]):
        self.clusters = clusters

    def as_json(self) -> str:
        return json.dumps(self.clusters)

    def as_table(self) -> Table:
        table = Table("id", "name", "type", "status", "created", show_header=True, header_style="bold green")
        phases = {
            V1ClusterState.QUEUED: Text("queued", style="bold yellow"),
            V1ClusterState.PENDING: Text("pending", style="bold yellow"),
            V1ClusterState.RUNNING: Text("running", style="bold green"),
            V1ClusterState.FAILED: Text("failed", style="bold red"),
            V1ClusterState.DELETED: Text("deleted", style="bold red"),
        }

        cluster_type_lookup = {
            V1ClusterType.BYOC: Text("byoc", style="bold yellow"),
            V1ClusterType.GLOBAL: Text("grid-cloud", style="bold green"),
        }
        for cluster in self.clusters:
            cluster: Externalv1Cluster
            status = phases[cluster.status.phase]
            if cluster.spec.desired_state == V1ClusterState.DELETED \
                    and cluster.status.phase != V1ClusterState.DELETED:
                status = Text("terminating", style="bold red")

            # this guard is necessary only until 0.3.93 releases which includes the `created_at`
            # field to the external API
            created_at = datetime.now()
            if hasattr(cluster, "created_at"):
                created_at = cluster.created_at

            table.add_row(
                cluster.id,
                cluster.name,
                cluster_type_lookup.get(cluster.spec.cluster_type, Text("unknown", style="red")),
                status,
                arrow.get(created_at).humanize() if created_at else "",
            )
        return table


@rich_click.group(invoke_without_command=True)
@click.pass_context
def clusters(ctx: click.Context) -> None:
    if ctx.invoked_subcommand is not None:
        return

    api_client = create_swagger_client()
    cluster_api = cluster_service_api.ClusterServiceApi(api_client)
    resp = cluster_api.cluster_service_list_clusters(phase_not_in=[V1ClusterState.DELETED])
    print_to_console(ctx, ClusterList(resp.clusters if resp.clusters else []))


default_instance_types = [
    "g2.8xlarge",
    "g3.16xlarge",
    "g3.4xlarge",
    "g3.8xlarge",
    "g3s.xlarge",
    "g4dn.12xlarge",
    "g4dn.16xlarge",
    "g4dn.2xlarge",
    "g4dn.4xlarge",
    "g4dn.8xlarge",
    "g4dn.metal",
    "g4dn.xlarge",
    "p2.16xlarge",
    "p2.8xlarge",
    "p2.xlarge",
    "p3.16xlarge",
    "p3.2xlarge",
    "p3.8xlarge",
    "p3dn.24xlarge",
    # "p4d.24xlarge",  # currently not supported
    "t2.large",
    "t2.medium",
    "t2.xlarge",
    "t2.2xlarge",
    "t3.large",
    "t3.medium",
    "t3.xlarge",
    "t3.2xlarge",
]


def _check_cluster_name_is_valid(_ctx, _param, value):
    pattern = r"^(?!-)[a-z0-9-]{1,63}(?<!-)$"
    if not re.match(pattern, value):
        raise click.ClickException(
            f"cluster name doesn't match regex pattern {pattern}\nIn simple words, use lowercase letters, numbers, and occasional -"
        )
    return value


# https://docs.grid.ai/platform/known-issues#byoc
UNSUPPORTED_ARCHITECTURES = ["a1", "t4g", "m6g", "c6g", "g5g", "r6g"]


def assert_instance_types_are_supported(instance_types: List[str]):
    # Checking for unsupported architectures in instance types
    #   https://docs.grid.ai/platform/known-issues#byoc

    unsupported = [
        instance_type for instance_type in instance_types
        if any(instance_type.startswith(arch) for arch in UNSUPPORTED_ARCHITECTURES)
    ]

    if unsupported:
        raise click.ClickException(
            f"Instance types {', '.join(unsupported)} are not supported due to their architecture. "
            f"Currently we do not support following architectures {', '.join(UNSUPPORTED_ARCHITECTURES)}. "
            f"Please choose supported instance type."
        )


@clusters.command()
@click.argument('name', type=str, callback=_check_cluster_name_is_valid)
@click.option('--external-id', 'external_id', type=str, required=True)
@click.option(
    '--role-arn', 'role_arn', type=str, required=True, help="AWS role ARN attached to`the associated resources."
)
@click.option(
    '--region',
    'region',
    type=str,
    required=False,
    default="us-east-1",
    help="AWS region which is used to host the associated resources."
)
@click.option(
    '--instance-types',
    'instance_types',
    type=str,
    required=False,
    default=",".join(default_instance_types),
    help="Instance types which you desire to support for computer jobs within the cluster."
)
@click.option(
    '--cost-savings',
    'cost_savings',
    type=bool,
    required=False,
    default=False,
    is_flag=True,
    help='using this flag ensures that the cluster is created with a profile that is optimized for '
    'cost saving, making runs cheaper but start-up times may increase',
)
@click.option(
    '--wait',
    'wait',
    type=bool,
    required=False,
    default=False,
    is_flag=True,
    help='using this flag CLI will wait until the cluster is running',
)
@click.option(
    '--edit-before-creation',
    default=False,
    is_flag=True,
    help="Edit the created cluster spec before submitting to API server."
)
def aws(
    name: str,
    external_id: str,
    role_arn: str,
    region: str,
    instance_types: str,
    edit_before_creation: bool,
    cost_savings: bool,
    wait: bool,
) -> None:
    """Create a grid compute cluster with NAME from the provided AWS account details."""
    performance_profile = V1ClusterPerformanceProfile.DEFAULT
    if cost_savings:
        performance_profile = V1ClusterPerformanceProfile.COST_SAVING

    api_client = create_swagger_client()
    cluster_api = cluster_service_api.ClusterServiceApi(api_client)
    driver = V1ClusterDriver(
        kubernetes=V1KubernetesClusterDriver(
            aws=V1AWSClusterDriverSpec(
                region=region,
                role_arn=role_arn,
                external_id=external_id,
                instance_types=[V1InstanceSpec(name=x) for x in instance_types.split(",")]
            )
        )
    )
    spec = V1ClusterSpec(cluster_type=V1ClusterType.BYOC, performance_profile=performance_profile, driver=driver)
    body = V1CreateClusterRequest(name=name, spec=spec)

    new_body = body
    if edit_before_creation:
        after = click.edit(json.dumps(body.to_dict(), indent=4))
        if after is not None:
            new_body = create_openapi_object(string2dict(after), body)
        if new_body == body:
            click.echo("cluster unchanged")

    # Check if instance types are supported -> Throws Excetpion
    # instance_types = new_body.spec.driver.kubernetes.aws.instance_types
    assert_instance_types_are_supported([i.name for i in new_body.spec.driver.kubernetes.aws.instance_types])

    resp = cluster_api.cluster_service_create_cluster(body=new_body)

    if wait:
        api_client = create_swagger_client()
        cluster_api = cluster_service_api.ClusterServiceApi(api_client)
        start = time.time()
        elapsed = 0
        while elapsed < MAX_CLUSTER_WAIT_TIME:
            cluster_resp = cluster_api.cluster_service_list_clusters(phase_not_in=[V1ClusterState.DELETED])
            new_cluster = None
            for clust in cluster_resp.clusters:
                if clust.id == resp.id:
                    new_cluster = clust
                    break
            if new_cluster is not None:
                if new_cluster.status.phase == V1ClusterState.RUNNING:
                    break
                elif new_cluster.status.phase == V1ClusterState.FAILED:
                    raise click.ClickException(f"Creation failed for cluster {resp.id}")
                time.sleep(CLUSTER_STATE_CHECKING_TIMEOUT)
            elapsed = time.time() - start
        else:
            raise click.ClickException(f"Max time for cluster creation is elapsed")

    click.echo(resp.to_str())


# TODO(nmiculinic): refactor this to be under `grid logs`
# Though I don't want to break existing workflows...
@clusters.command()
@click.argument('cluster_id', type=str, callback=_check_cluster_name_is_valid)
@click.option("-t", "--tail", is_flag=True, default=False, help="whether to tail log lines")
@click.option(
    "--from",
    "from_time",
    default="24 hours ago",
    help="The starting timestamp to query cluster logs from.",
    callback=arrow_time_callback
)
@click.option(
    "--to",
    "to_time",
    default="0 seconds ago",
    callback=arrow_time_callback,
    help="The end timestamp / relative time increment to query logs for. This is ignored when tailing logs."
)
@click.option("--limit", default=1000, help="The max number of log lines returned.")
@click.option(
    "--time-format",
    default="iso8601",
    type=click.Choice(['human', 'iso8601'], case_sensitive=False),
    help="Timestamp formatting style",
)
def logs(time_format: str, limit: int, to_time: arrow.Arrow, from_time: arrow.Arrow, tail: bool, cluster_id: str):
    """Retrieve cluster logs from the managed cluster identified by CLUSTER_ID.

    These logs are streamed to stdout, and can either be tailed to view log
    lines as they are generated, or limited to a time range.
    """
    ws_client = create_websocket_client()
    cluster_service = ClusterService(ws_client)
    req = ClusterLogsRequest(
        cluster_id=cluster_id, follow=tail, start=from_time.int_timestamp, end=to_time.int_timestamp, limit=limit
    )
    for log_entry in cluster_service.cluster_logs(req):
        timestamp = arrow.get(log_entry.timestamp)
        if time_format == "human":
            timestamp_part = click.style("{:>15}".format(timestamp.humanize()), fg='green')
        elif time_format == "iso8601":
            timestamp_part = click.style(timestamp.format("YYYY-MM-DD HH:mm:ss"), fg='green')
        else:
            raise click.ClickException(f"Unknown time format: got {time_format}, valid entries: [human,iso8601]")
        log_level = click.style(log_entry.level, fg='magenta')
        message_part = log_entry.message.rstrip()
        click.echo(f"[{timestamp_part}] [{log_level}]: {message_part}")



from datetime import datetime
import textwrap
from pathlib import Path
from enum import Enum
from typing import Optional, Union, List

from grid.openapi import V1SessionSpec, V1SessionState, V1SessionStatus, Externalv1Session
from grid.sdk import env
from grid.sdk.client import create_swagger_client
from grid.sdk.datastores import Datastore
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.sessions import list_sessions as rest_list_sessions, get_session_from_name, get_session_from_id
from grid.sdk.rest.datastores import datastore_id_from_dsn
from grid.sdk.rest.sessions import (
    change_session_instance_type,
    create_session,
    delete_session,
    pause_session,
    resume_session,
)
from grid.sdk.affirmations import affirm, is_not_deleted, is_created, is_not_created, is_not_shallow
from grid.sdk.user import get_teams, User, user_from_logged_in_account
from grid.sdk.utilities import SPECIAL_NAME_TO_SKIP_OBJECT_INIT, read_config
from grid.sdk.utils import fill_object_from_response
from grid.sdk.utils.name_generator import unique_name


class Session:
    """Specify, Modify, Create, Pause, or Delete an interactive session instance.

    Interactive sessions are optimized for development activites (before executing
    hyperparemeter sweeps in a Run). Once created, sessions can be accessed via
    VSCode, Jupyter-lab, or SSH interfaces.

    Grid manages the installation of any/all core libraries, drivers, and interfaces
    to the outside world. Sessions can be run on anything from a small 2 CPU core +
    4GB memory instance to a monster machine with 96 CPU cores + 824 GB memory + eight
    V100 GPUs + 40 GBPS network bandwidth (no, those values aren't typos!); or really
    anything in between.

    Parameters
    ----------
    name
        human addressable name to assign to the session instance
    instance_type
        compute node type which the session will be provisioned on
    disk_size_gb
        amount of storage provisioned to the root boot disk the session
        will be provisioned on (ie. disk storage provisioned to the user)
    datastore
        datastore object
    datastore_mount_dir
        directory on the session to mount the provided datastore
    use_spot
        bool indicating if a spot instance should be used when provisioining
        the underlying machine the session service operates on. If True
        then the hourly cost will be significantly reduced from the base
        (on-demand) price, but the instance can be terminated at any time.
        Upon termination your notebooks and scripts will be saved, but you may
        lose data which is "in-process".

        If False (the default), then un-interuptable on-demand instances are
        used. While this increases costs it does mean that your machine will
        not be deprovisioned if the cloud provider experiences increased
        demand for that instance type.
    cluster_id
        Bring-your-own-credential users only. specify the name of the cluster
        to operate the sesison on.
    """

    _name: str
    _id: str
    _cluster_id: str

    _datastore: Optional[Datastore]
    _datastore_mount_dir: Optional[str]
    _disk_size_gb: str
    _instance_type: str
    _use_spot: bool

    _hourly_cost: float
    _total_cost: float
    _total_run_time: float

    _desired_state: 'SessionState'
    _status: 'SessionState'

    _created_at: datetime
    _started_at: datetime
    _finished_at: datetime
    _last_state_status_transition_at: datetime

    _jupyter_lab_url: str
    _jupyter_lab_token: str
    _ssh_url: str

    _client: GridRestClient

    _user: User

    _is_deleted: bool
    _is_created: bool

    def __init__(
        self,
        name: str,
        instance_type: Optional[str] = None,
        disk_size_gb: Union[int, str] = 200,
        datastore: Optional[Datastore] = None,
        datastore_mount_dir: Optional[str] = None,
        use_spot: bool = False,
        cluster_id: Optional[str] = None,
        config_file: Optional[Union[str, Path]] = None,
    ):
        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This block sets the correct values for the private      #
        #    attributes which is then later picked up by the decorator(s) to take     #
        #    right actions. It also initialize the _client object and cluster_id      #
        #    which is required by the downstream methods regardless of the object     #
        #    is completely initialized or not. Most importantly, this blocks checks   #
        #    for the name argument to decide if it's a call from other internal       #
        #    methods to create a shallow object. This is done by checking the         #
        #    special name variable. Other methods that already has the backend        #
        #    response fetched, can use this to create the object without the backend  #
        #    call and then fill-in the response they already have.                    #
        #                                                                             #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id
        if name == SPECIAL_NAME_TO_SKIP_OBJECT_INIT:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #
        if name is None:
            name = unique_name()
        else:
            try:
                session = get_session_from_name(self._client, name, cluster_id)
                self._setup_from_response(session)
                return
            except KeyError:
                self._is_deleted = False  # the run has not been deleted
                self._is_created = False  # it doesn't exists in the grid backend.
                pass

        if not datastore_mount_dir and isinstance(datastore, Datastore):
            datastore_mount_dir = f'/datastores/{datastore.name}'

        # The config option should be deprecated. This just takes the file passed in an
        # attempts to get the values provided within it, defaulting back to the standard
        # CLI click option values if they are not provided.
        if config_file is not None:
            config = read_config(config_file)
            cluster_id = config.get('compute', {}).get('provider', {}).get('cluster', cluster_id)
            instance_type = config.get('compute', {}).get('train').get('instance_type', instance_type)
            disk_size_gb = config.get('compute', {}).get('train', {}).get('disk_size', disk_size_gb)
            use_spot = config.get('compute', {}).get('train', {}).get('use_spot', use_spot)
            datastore_name = config.get('compute', {}).get('train', {}).get('datastore_name')
            datastore_version = config.get('compute', {}).get('train', {}).get('datastore_version')
            datastore_mount_dir = config.get('compute', {}).get('train',
                                                                {}).get('datastore_mount_dir', datastore_mount_dir)
            _dstore_config = [datastore_name, datastore_version]
            if any(_dstore_config) and not all(_dstore_config):
                raise ValueError(
                    "name and version are both required if one of datastore_name or datastore_version is set"
                )
            if all(_dstore_config):
                datastore = Datastore(name=datastore_name, version=datastore_version)

        self._name = name
        self._instance_type = instance_type
        self._disk_size_gb = str(disk_size_gb)

        self._datastore = datastore
        self._datastore_mount_dir = datastore_mount_dir

        self._use_spot = use_spot
        self._cluster_id = cluster_id
        self._name = name
        self._user = user_from_logged_in_account()

        self._is_deleted = False
        self._is_created = False
        self._has_initialized = True

    @classmethod
    def _from_id(cls, session_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Session":
        instance = cls(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT, cluster_id=cluster_id)
        instance._id = session_id
        instance._is_shallow = True
        return instance

    def _unshallow(self):
        """ If the object is a shallow (i.e. only has an id and `_is_shallow` attribute is True)
        object, this method can be triggered to get the full object from the BE. It is designed
        to be called only from the `is_not_shallow` decorator and should not be called directly.
        """
        if not self._is_shallow:
            raise RuntimeError('Session is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid session id")
        self._is_shallow = False
        run = get_session_from_id(self._client, run_id=self._id, cluster_id=self.cluster_id)
        self._setup_from_response(run)

    @affirm(is_not_shallow, is_not_deleted)
    def __repr__(self):
        if self._is_created is True:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <17} = \"{self.name}\",
                    {"cluster_id": <17} = \"{self.cluster_id}\",
                    {"datastore": <17} = {self.datastore},
                    {"disk_size_gb": <17} = {self.disk_size_gb},
                    {"instance_type": <17} = \"{self.instance_type}\",
                    {"use_spot": <17} = {self.use_spot},
                    {"hourly_cost": <17} = {self.hourly_cost},
                    {"total_cost": <17} = {self.total_cost},
                    {"total_run_time": <17} = {self.total_run_time},
                    {"desired_state": <17} = {self.desired_state.value},
                    {"last_status_state": <17} = {self.status.value},
                    {"jupyter_lab_url": <17} = \"{self.jupyter_lab_url}\",
                    {"jupyter_lab_token": <17} = \"{self.jupyter_lab_token}\",
                    {"ssh_url": <17} = \"{self.ssh_url}\",
                )"""
            )
        else:
            res = textwrap.dedent(
                f"""\
            {self.__class__.__name__}(
                {"name": <17} = \"{self.name}\",
                {"cluster_id": <17} = \"{self.cluster_id}\",
                {"datastore": <17} = {self.datastore},
                {"disk_size_gb": <17} = {self.disk_size_gb},
                {"instance_type": <17} = \"{self.instance_type}\",
                {"use_spot": <17} = {self.use_spot},
            )"""
            )
        return res

    # ---------------- User Modifiable Attributes -----------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """The name of the session."""
        return self._name

    @name.setter
    @affirm(is_not_created)
    def name(self, value: str):
        self._name = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        """The name of the cluster the session is running on.
        """
        return self._cluster_id

    @cluster_id.setter
    @affirm(is_not_shallow, is_created)
    def cluster_id(self, value: str):
        self._cluster_id = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore(self) -> Datastore:
        return self._datastore

    @datastore.setter
    @affirm(is_not_shallow, is_created)
    def datastore(self, value: Datastore):
        self._datastore = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore_mount_dir(self) -> str:
        return self._datastore_mount_dir

    @datastore_mount_dir.setter
    @affirm(is_not_shallow, is_created)
    def datastore_mount_dir(self, value: str):
        self._datastore_mount_dir = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def disk_size_gb(self) -> str:
        """The size of the home directory disk to spin up
        """
        return self._disk_size_gb

    @disk_size_gb.setter
    @affirm(is_not_shallow, is_created)
    def disk_size_gb(self, value: str):
        self._disk_size_gb = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def instance_type(self) -> str:
        """The type of the virtual machine used on the compute cluster when running the session.
        """
        return self._instance_type

    @instance_type.setter
    @affirm(is_not_shallow, is_created)
    def instance_type(self, value: str):
        self._instance_type = value

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def use_spot(self) -> bool:
        """If a spot instance type is used to spin up the session.
        """
        return self._use_spot

    @use_spot.setter
    @affirm(is_not_shallow, is_created)
    def use_spot(self, value: bool):
        self._use_spot = value

    # ----------------------- Fixed Attributes ---------------------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def hourly_cost(self) -> float:
        """The per hour cost of this session configuration when it is in a 'RUNNING' state.
        """
        return self._hourly_cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def total_cost(self) -> float:
        """The total cost of the session over it's entire lifetime.
        """
        return self._total_cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def total_run_time(self) -> float:
        """How long the session has run for (not including paused time) in second.
        """
        return self._total_run_time

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def desired_state(self) -> 'SessionState':
        """The state of the system we are trying to achieve.

        This might be one of 'PENDING', 'RUNNING', 'PAUSED'
        """
        return self._desired_state

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def status(self) -> 'SessionState':
        """The last recorded state of the session instance on the grid platform.

        This might be one of 'PENDING', 'RUNNING', 'PAUSED'
        """
        return self._status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        """The timestamp when the session was first created.
        """
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def started_at(self) -> datetime:
        """The timestamp when the session was last started.
        """
        return self._started_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def finished_at(self) -> datetime:
        """The timestamp when the session was last stopped.
        """
        return self._finished_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def last_state_status_transition_at(self) -> datetime:
        """The last timestamp when the session's running state was changed.
        """
        return self._last_state_status_transition_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def jupyter_lab_url(self) -> str:
        """URL to access the jupyterlab server at over the public internet.
        """
        return self._jupyter_lab_url

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def jupyter_lab_token(self) -> str:
        """Security token required to access this session over the public internet.
        """
        return self._jupyter_lab_token

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def ssh_url(self) -> str:
        """URL used to SSH into this session.
        """
        return self._ssh_url

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        """Details of the user who is the creator of this session.
        """
        return self._user

    @property
    def exists(self) -> bool:
        """If this object refers to a session which has been created on the grid platform.

        Prior to calling `start` for the frist time, the grid platform has not created
        the actual session implementation, and the object does not exist.
        """
        return self._is_created

    # ---------------------- Interaction Methods -----------------------------------

    def start(self) -> bool:
        """Start an interactive session based on this configuration.

        If the session does not exist, a new session will be created; if the session
        exists, but is paused, then the session will be resumed; if the session exists
        and is already running, no action will be taken.
        """
        if self._is_created is True:
            session = resume_session(self._client, session_id=self._id, cluster_id=self._cluster_id)
        else:
            datastores = {self.datastore.id: self.datastore_mount_dir} if self.datastore else None
            session = create_session(
                self._client,
                name=self._name,
                instance_type=self._instance_type,
                cluster_id=self._cluster_id,
                datastores=datastores,
                disk_size_gb=self._disk_size_gb,
                use_spot=self._use_spot
            )
        self._setup_from_response(session)
        return True

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def pause(self) -> bool:
        """Pauses a session which is currently running.

        Pausing a session stops the running instance (and any computations being
        performed on it - be sure to save your work!) and and billing of your account
        for the machine. The session can be resumed at a later point with all your
        persisted files and saved work unchanged.
        """
        session = pause_session(self._client, session_id=self._id, cluster_id=self._cluster_id)
        self._setup_from_response(session)
        return True

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def delete(self) -> bool:
        """Deletes a session which is either running or paused.

        Deleting a session will stop the running instance (and any computations being
        performed on it) and billing of your account. All work done on the machine
        is permenantly removed, including all/any saved files, code, or downloaded
        data (assuming the source of the data was not a grid datastore - datastore
        data is not deleted).
        """
        delete_session(self._client, session_id=self._id, cluster_id=self._cluster_id)
        self._is_deleted = True
        return True

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def change_instance_type(self, instance_type: str, use_spot: Optional[bool] = None) -> bool:
        """Change the instance type of a session.

        The session must be paused before calling this method.

        Parameters
        ----------
        instance_type
            the new instance type the session node should use.
        use_spot
            if true, use interuptable spot instances (which come at a steap discount,
            but which can be interrupted and shut down at any point in time depending
            on cloud provider instance type demand). If false, use an on-demand instance.

            By default this value is None, indicating that no change will be made to the
            current configuratoin.
        """
        if self.status != SessionState.PAUSED:
            raise RuntimeError("session must be paused before calling `change_instance_type`.")

        resp = change_session_instance_type(
            self._client,
            session_id=self._id,
            instance_type=instance_type,
            use_spot=use_spot,
            cluster_id=self._cluster_id
        )
        self._setup_from_response(resp)
        return True

    def _setup_from_response(self, session: Externalv1Session):
        self._is_created = True
        self._is_deleted = session.status.phase == V1SessionState.DELETED
        self._is_shallow = False

        spec: V1SessionSpec = session.spec
        status: V1SessionStatus = session.status

        self._cluster_id = spec.cluster_id
        self._id = session.id
        self._name = session.name
        self._instance_type = spec.instance_type
        self._disk_size_gb = spec.resources.storage_gb
        self._use_spot = spec.use_spot
        self._datastore = None
        self._datastore_mount_dir = None
        for dstore_input in spec.datastores:
            datastore_id = datastore_id_from_dsn(dstore_input.dsn)
            # noinspection PyProtectedMember
            datastore = Datastore._from_id(datastore_id, cluster_id=self.cluster_id)
            datastore._id = datastore_id
            self._datastore = datastore
            self._datastore_mount_dir = dstore_input.mount_path

        # TODO - getattr because BE is sometimes respond with different types. Unify this
        self._hourly_cost = getattr(session, "hourly_cost", None)
        self._total_cost = getattr(session, "cost", None)
        self._total_run_time = float(status.total_run_time_seconds or 0)

        self._desired_state = SessionState.from_api_spec(spec.desired_state)
        self._status = SessionState.from_api_spec(status.phase)

        self._created_at = session.created_at
        self._started_at = status.start_timestamp
        self._finished_at = status.stop_timestamp
        self._last_state_status_transition_at = status.last_state_status_transition_timestamp

        self._jupyter_lab_url = status.jupyter_lab_url
        self._jupyter_lab_token = status.jupyter_lab_token
        self._ssh_url = status.ssh_url

        # TODO: complete with owner's username, first name, & last name
        self._user = User(user_id=session.spec.user_id, username="", first_name="", last_name="")


def list_sessions(cluster_id: Optional[str] = None, include_teams: Optional[bool] = False) -> List[Session]:
    """List sessions for user/team

    Parameters
    ----------
    cluster_id:
        the cluster id to list sessions from
    include_teams:
        if True, returns a list of sessions of the everyone in the team

    Returns
    -------
    List[Session]
        sequence of session interaction objects.
    """
    cluster_id = cluster_id or env.CONTEXT
    c = GridRestClient(create_swagger_client())
    users = {}
    user_ids = None
    if include_teams is True:
        teams = get_teams()
        user_ids = []
        for team in teams.values():
            for user_id in team.members:
                users[user_id] = team.members[user_id]
                user_ids.append(user_id)
    sessions = []
    for session_def in rest_list_sessions(c, user_ids=user_ids, cluster_id=cluster_id):
        session_obj = Session(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT)
        fill_object_from_response(session_obj, session_def)
        session_obj._user = users.get(session_obj.user.user_id, session_obj.user)
        sessions.append(session_obj)
    return sessions


class SessionState(Enum):
    UNSPECIFIED = "unspecified"
    PENDING = "pending"
    RUNNING = "running"
    PAUSED = "paused"
    DELETED = "deleted"
    FAILED = "failed"

    @classmethod
    def from_api_spec(cls, status: V1SessionState) -> 'SessionState':
        parsed = str(status).lower().split('_', maxsplit=2)[-1]
        return cls(parsed)



import datetime
import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, List

from grid.sdk import env
from grid.sdk._gql.queries import get_user_basic_info, get_user_teams


@dataclass
class User:
    user_id: str
    username: str
    first_name: str
    last_name: str
    email: Optional[str] = None
    is_verified: Optional[bool] = None
    is_blocked: Optional[bool] = None
    completed_signup: Optional[bool] = None
    team_name: Optional[str] = None


def set_default_cluster(cluster_id):
    """Set the default cluster in the local configuration
    and in the backend (backend is TODO)
    """
    settings_path = Path(env.GRID_SETTINGS_PATH)
    if not settings_path.exists():
        env.write_default_settings(settings_path)
    user_settings = json.load(settings_path.open())
    user_settings['context'] = cluster_id
    with settings_path.open('w') as file:
        json.dump(user_settings, file, ensure_ascii=False, indent=4)
    os.environ['GRID_CLUSTER_ID'] = cluster_id
    env.reset_global_variables()


def user_from_logged_in_account() -> "User":
    resp = get_user_basic_info()
    if not resp["isVerified"]:
        raise PermissionError(
            f"User account not yet verified. Verify your "
            f"account at {env.GRID_URL}/#/verification"
        )
    if not resp["completedSignup"]:
        raise PermissionError(
            f"You haven't yet completed registration. Please complete "
            f"registration at {env.GRID_URL}/#/registration"
        )
    if resp["isBlocked"]:
        raise PermissionError(
            f"Your account with username `{resp['username']}` has been "
            f"suspended. Please reach out to support at support@grid.ai"
        )

    return User(
        user_id=resp["userId"],
        username=resp["username"],
        first_name=resp["firstName"],
        last_name=resp["lastName"],
        email=resp["email"],
        is_verified=resp["isVerified"],
        is_blocked=resp["isBlocked"],
        completed_signup=resp["completedSignup"],
    )


@dataclass(frozen=True)
class Team:
    team_id: str
    name: str
    created_at: datetime.datetime
    role: str
    members: Dict[str, User]


def get_teams() -> Dict[str, Team]:
    team_definitions = get_user_teams()
    teams = {}
    for team in team_definitions:
        members = {}
        for member in team['members']:
            user = User(
                user_id=member['id'],
                username=member['username'],
                first_name=member['firstName'],
                last_name=member['lastName'],
            )
            members[user.user_id] = user

        team = Team(
            team_id=team['id'],
            name=team['name'],
            created_at=datetime.datetime.fromisoformat(team['createdAt']),
            role=team['role'],
            members=members,
        )
        teams[team.team_id] = team

    return teams


def get_user_team_members() -> List[User]:
    """Return all users (id, username, team name) of the currently logged in user"""
    users = []
    for team_data in get_user_teams():
        for member_data in team_data['members']:
            users.append(
                User(
                    user_id=member_data['id'],
                    username=member_data['username'],
                    team_name=team_data['name'],
                    first_name=member_data['firstName'],
                    last_name=member_data['lastName']
                )
            )
    return users



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1AzureClusterDriverSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'instance_types': 'list[V1InstanceSpec]',
        'internal_test_cluster': 'bool',
        'kubernetes_version': 'str',
        'region': 'str',
        'subscription_id': 'str'
    }

    attribute_map = {
        'instance_types': 'instanceTypes',
        'internal_test_cluster': 'internalTestCluster',
        'kubernetes_version': 'kubernetesVersion',
        'region': 'region',
        'subscription_id': 'subscriptionId'
    }

    def __init__(self, instance_types: 'list[V1InstanceSpec]' = None, internal_test_cluster: 'bool' = None, kubernetes_version: 'str' = None, region: 'str' = None, subscription_id: 'str' = None, _configuration=None):  # noqa: E501
        """V1AzureClusterDriverSpec - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._instance_types = None
        self._internal_test_cluster = None
        self._kubernetes_version = None
        self._region = None
        self._subscription_id = None
        self.discriminator = None

        if instance_types is not None:
            self.instance_types = instance_types
        if internal_test_cluster is not None:
            self.internal_test_cluster = internal_test_cluster
        if kubernetes_version is not None:
            self.kubernetes_version = kubernetes_version
        if region is not None:
            self.region = region
        if subscription_id is not None:
            self.subscription_id = subscription_id

    @property
    def instance_types(self) -> 'list[V1InstanceSpec]':
        """Gets the instance_types of this V1AzureClusterDriverSpec.  # noqa: E501


        :return: The instance_types of this V1AzureClusterDriverSpec.  # noqa: E501
        :rtype: list[V1InstanceSpec]
        """
        return self._instance_types

    @instance_types.setter
    def instance_types(self, instance_types: 'list[V1InstanceSpec]'):
        """Sets the instance_types of this V1AzureClusterDriverSpec.


        :param instance_types: The instance_types of this V1AzureClusterDriverSpec.  # noqa: E501
        :type: list[V1InstanceSpec]
        """

        self._instance_types = instance_types

    @property
    def internal_test_cluster(self) -> 'bool':
        """Gets the internal_test_cluster of this V1AzureClusterDriverSpec.  # noqa: E501


        :return: The internal_test_cluster of this V1AzureClusterDriverSpec.  # noqa: E501
        :rtype: bool
        """
        return self._internal_test_cluster

    @internal_test_cluster.setter
    def internal_test_cluster(self, internal_test_cluster: 'bool'):
        """Sets the internal_test_cluster of this V1AzureClusterDriverSpec.


        :param internal_test_cluster: The internal_test_cluster of this V1AzureClusterDriverSpec.  # noqa: E501
        :type: bool
        """

        self._internal_test_cluster = internal_test_cluster

    @property
    def kubernetes_version(self) -> 'str':
        """Gets the kubernetes_version of this V1AzureClusterDriverSpec.  # noqa: E501


        :return: The kubernetes_version of this V1AzureClusterDriverSpec.  # noqa: E501
        :rtype: str
        """
        return self._kubernetes_version

    @kubernetes_version.setter
    def kubernetes_version(self, kubernetes_version: 'str'):
        """Sets the kubernetes_version of this V1AzureClusterDriverSpec.


        :param kubernetes_version: The kubernetes_version of this V1AzureClusterDriverSpec.  # noqa: E501
        :type: str
        """

        self._kubernetes_version = kubernetes_version

    @property
    def region(self) -> 'str':
        """Gets the region of this V1AzureClusterDriverSpec.  # noqa: E501


        :return: The region of this V1AzureClusterDriverSpec.  # noqa: E501
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region: 'str'):
        """Sets the region of this V1AzureClusterDriverSpec.


        :param region: The region of this V1AzureClusterDriverSpec.  # noqa: E501
        :type: str
        """

        self._region = region

    @property
    def subscription_id(self) -> 'str':
        """Gets the subscription_id of this V1AzureClusterDriverSpec.  # noqa: E501


        :return: The subscription_id of this V1AzureClusterDriverSpec.  # noqa: E501
        :rtype: str
        """
        return self._subscription_id

    @subscription_id.setter
    def subscription_id(self, subscription_id: 'str'):
        """Sets the subscription_id of this V1AzureClusterDriverSpec.


        :param subscription_id: The subscription_id of this V1AzureClusterDriverSpec.  # noqa: E501
        :type: str
        """

        self._subscription_id = subscription_id

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1AzureClusterDriverSpec, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1AzureClusterDriverSpec') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1AzureClusterDriverSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1AzureClusterDriverSpec') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1AzureClusterDriverSpec):
            return True

        return self.to_dict() != other.to_dict()



from typing import Optional

import click

from grid.cli import rich_click
from grid.cli.client import Grid


@rich_click.command('sync-env', cls=rich_click.deprecate_grid_options())
@click.option('--config', 'config_path', required=False, help='Path to Grid config YML')
@click.option(
    '--dependency_file',
    required=False,
    help='Path to dependency file. Defaults to the '
    '`requirements.txt` or `environment.yml` found in the root'
)
def sync_env(config_path: Optional[str] = None, dependency_file: Optional[str] = None):
    """
    Synchronize the requirements file with packages and versions
    from the currently active environment
    """
    client = Grid()
    req_file = client.serialize_dependencies(config_path, dependency_file)
    if req_file:
        message = f"""
Environment has been synced with requirements file. Make sure to
commit the file before running the experiments

        git add {req_file}
        git commit -m "{req_file} synced with current environment"
        """
        click.echo(message)



from grid.openapi import V1ClusterState, V1ClusterType, V1ListClustersResponse
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message


@throw_with_message
def list_clusters(c: GridRestClient, is_global: bool = False) -> V1ListClustersResponse:
    if is_global:
        resp = c.cluster_service_list_clusters(
            phase_not_in=[V1ClusterState.DELETED], cluster_type_in=[V1ClusterType.GLOBAL, V1ClusterType.BYOC]
        )
    else:
        resp = c.cluster_service_list_clusters(phase_not_in=[V1ClusterState.DELETED])
    return resp



from datetime import datetime, timezone
from typing import Optional, List

import click
from rich.console import Console
from rich.table import Table
from yaspin import yaspin

from grid.cli import rich_click
from grid.cli.client import Grid
from grid.cli.observables.base import create_table
from grid.cli.types import ObservableType
from grid.cli.utilities import is_experiment, get_abs_time_difference, string_format_timedelta, get_param_values, \
    get_experiment_duration_string, get_experiment_queued_duration_string
from grid.sdk import env
from grid.sdk.experiments import Experiment, ExperimentState
from grid.sdk.rest.exceptions import GridException
from grid.sdk.runs import list_runs, Run, RunState


def _check_is_experiment(ctx, _param, value):
    """Callback that checks if a value is an experiment."""
    if not value:
        return value

    if is_experiment(value):
        raise click.BadArgumentUsage(f"Must pass a Run name, not Experiment: {value}")

    return value


@rich_click.command()
@rich_click.argument('run', type=str, nargs=1, required=False, callback=_check_is_experiment)
@click.option(
    '--global', 'is_global', type=bool, is_flag=True, help='Fetch status from all collaborators when flag is passed'
)
def status(run: Optional[str] = None, is_global: Optional[bool] = False) -> None:
    """Checks the status of Runs, Experiments, and Sessions."""

    click.echo(f"Loading status in {env.CONTEXT}")
    # Initializing variables
    spinner_load_type = "Runs"
    history_runs = []
    if run:
        spinner_load_type = "Experiments"
    spinner = yaspin(text=f"Loading {spinner_load_type} in {env.CONTEXT}...", color="yellow")
    console = Console()

    # Fetching runs/experiments and printing them onto the terminal
    try:
        spinner.start()
        runs = list_runs(is_global=is_global, query=run)
        if not run:
            table = render_runs(runs, is_global=is_global)
            history_runs = len(runs) - len(table.rows)
        else:
            for run_obj in runs:
                if run_obj.name == run:
                    if len(run_obj.experiments) == 0:
                        raise RuntimeError(f"Run {run} has no experiments")
                    # TODO - this is working because of the order which is_unshallow
                    #   and is_delted being called. We should deprecate is_deleted and
                    #   instead use _status on all resources
                    non_deleted_exps = [exp for exp in run_obj.experiments if exp.status != ExperimentState.DELETED]
                    table = render_experiments(non_deleted_exps)
                    break
            else:
                # if break hasn't been called, then the run wasn't found
                raise RuntimeError(
                    f"Run {run} not found in the cluster {env.CONTEXT}. If you are "
                    f"looking for a run in a different cluster, try setting the default "
                    f"cluster with `grid user set_default_cluster <cluster-name>` first"
                )

        spinner.text = 'Done!'
        spinner.ok("✔")
        console.print(table)
    except GridException as e:
        spinner.fail("✘")
        raise click.ClickException(str(e))
    except Exception as e:
        spinner.fail("✘")
        raise click.ClickException(str(e))
    finally:
        spinner.stop()

    # Optionally fetching sessions, if run name is not passed
    if not run:
        if history_runs > 0:
            click.echo(f'{history_runs} Run(s) are not active. Use `grid history` ' 'to view your Run history.\n')
        client = Grid()
        client.status(kind=ObservableType.INTERACTIVE, identifiers=[], is_global=is_global)


def render_runs(runs: List['Run'], is_global: bool) -> Table:
    if is_global:
        table_cols = [
            'Run',
            'Status',
            'Created By',
            'Duration',
            'Experiments',
            'Running',
            'Queued',
            'Completed',
            'Failed',
            'Stopped',
        ]
    else:
        table_cols = [
            'Run',
            'Status',
            'Duration',
            'Experiments',
            'Running',
            'Queued',
            'Completed',
            'Failed',
            'Stopped',
        ]

    table = create_table(table_cols)
    table_rows = 0
    for run in runs:
        # we only have 3 statuses for runs
        # running (if something is running)
        is_running = run.status == RunState.RUNNING

        # We classify queued and pending into queued
        n_queued = run.experiment_counts.get(RunState.PENDING.value, 0)
        n_running = run.experiment_counts.get(RunState.RUNNING.value, 0)
        n_failed = run.experiment_counts.get(RunState.FAILED.value, 0)
        n_succeeded = run.experiment_counts.get(RunState.SUCCEEDED.value, 0)
        n_cancelled = run.experiment_counts.get(RunState.CANCELLED.value, 0)

        # If anything is queued, the the status of the entire
        # run is queued. All other statuses are running in
        # all other conditions.
        if n_queued > 0:
            run_status = "queued"

        # If you have anything running (and nothing queued)
        # then, mark the run as running.
        elif is_running:
            run_status = "running"

        # If it doesn't match the conditions above, just
        # skip this row and add the row and put it in history.
        else:
            # don't render table because it should be in history
            continue

        # Calculate the duration column
        delta = get_abs_time_difference(datetime.now(timezone.utc), run.created_at)
        duration_str = string_format_timedelta(delta)

        if is_global:
            table.add_row(
                run.name, run_status, run.user.username, duration_str, str(len(run.experiments)), str(), str(n_queued),
                str(n_succeeded), str(n_failed), str(n_cancelled)
            )
        else:
            table.add_row(
                run.name, run_status, duration_str, str(len(run.experiments)), str(n_running), str(n_queued),
                str(n_succeeded), str(n_failed), str(n_cancelled)
            )
        #  Let's count how many rows have been added.
        table_rows += 1
    return table


def render_experiments(experiments: List['Experiment']) -> Table:
    table = create_table([
        'Experiment',
        'Status',
        'Queued Duration',
        'Run Duration',
        'Command',
    ])

    for experiment in experiments:
        # Split hparam vals
        command = " ".join(experiment.command)

        duration_str = get_experiment_duration_string(
            created_at=experiment.created_at,
            started_running_at=experiment.start_timestamp,
            finished_at=experiment.finish_timestamp
        )
        queued_duration_str = get_experiment_queued_duration_string(
            created_at=experiment.created_at,
            started_running_at=experiment.start_timestamp,
        )
        table.add_row(experiment.name, experiment.status.value, queued_duration_str, duration_str, command)
    return table



from typing import Optional, List

from grid.openapi import (
    V1Credential,
    V1CreateCredentialRequest,
    V1CredentialType,
)
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message


@throw_with_message
def create_credential(
    c: GridRestClient,
    cluster_id: str,
    credential_type: V1CredentialType,
    credential_value: str,
) -> V1Credential:
    """Create a credential
    """
    resp = c.credential_service_create_credential(
        credential_cluster_id=cluster_id,
        body=V1CreateCredentialRequest(
            credential=V1Credential(
                cluster_id=cluster_id,
                type=credential_type,
                credential=credential_value,
            )
        )
    )
    return resp


@throw_with_message
def list_credentials(
    c: GridRestClient,
    cluster_id: str,
    user_ids: Optional[List[str]] = None,
    credential_type_in: Optional[List[V1CredentialType]] = None,
    credential_type_not_in: Optional[List[V1CredentialType]] = None,
) -> List[V1Credential]:
    """List all credentials fitting the filter criteria.
    """
    credentials: List[V1Credential] = []

    kwargs = {}
    if user_ids is not None:
        kwargs['user_ids'] = user_ids
    if credential_type_in is not None:
        kwargs['credential_type_in'] = credential_type_in
    if credential_type_not_in is not None:
        kwargs['credential_type_not_in'] = credential_type_not_in

    # loop until we recieve all credentials (and not just the max page size limit enforced by the backend)
    done_once, continuation_token = False, ""
    while (done_once is False) and (continuation_token == ""):
        done_once = True
        if continuation_token != "":
            kwargs['page_token'] = continuation_token

        resp = c.credential_service_list_credentials(
            cluster_id=cluster_id,
            **kwargs,
        )
        continuation_token = resp.next_page_token
        credentials.extend(resp.credentials)

    return credentials


@throw_with_message
def delete_credential(
    c: GridRestClient,
    cluster_id: str,
    credential_id: str,
) -> None:
    """Delete a credential, raising exception on failure.
    """
    c.credential_service_delete_credential(
        id=credential_id,
        cluster_id=cluster_id,
    )



import click

from grid.cli import rich_click
from grid.cli.client import credentials_from_env
from grid.cli.core.team import Team
from grid.sdk._gql.queries import get_user_info
from grid.sdk import user as sdk_user, list_clusters, env


@rich_click.group(invoke_without_command=True)
@click.pass_context
def user(ctx):
    """Show the user information of the authorized user for this CLI instance."""
    if ctx.invoked_subcommand is not None:
        return

    creds = credentials_from_env()
    user_info = get_user_info()

    email = user_info['email']
    email = email if email is not None else "N/A"

    click.echo(f"Display name    : {user_info['firstName']} {user_info['lastName']}")
    click.echo(f"UserID          : {creds.user_id}")
    click.echo(f"Username        : {user_info['username']}")
    click.echo(f"Email           : {email}")
    click.echo(f"Cluster context : {env.CONTEXT}")

    teams = Team.get_all()
    if teams:
        click.echo("Teams:\n-")
        for t in teams:
            click.echo(f"  {t.name} - Role: {t.role}")


@user.command()
@click.argument('cluster_name', type=str)
def set_default_cluster(cluster_name: str):
    """Specify the default CLUSTER_ID which all operations should be run against.
    """
    # TODO - remove this in favour of set_cluster_context
    click.echo(
        "set-default-cluster command is deprecated and will be "
        "removed in a future version. Please use set-cluster-context instead."
    )
    cluster_resp = list_clusters()
    for cluster in cluster_resp.clusters:
        if cluster_name == cluster.name:
            cluster_id = cluster.id
            break
    else:
        raise click.ClickException(f"Cluster with name {cluster_name} not found")
    sdk_user.set_default_cluster(cluster_id)
    click.echo(f"✔ Successfully set the cluster context to {cluster_name}")


@user.command()
@click.argument('cluster_name', type=str)
def set_cluster_context(cluster_name: str):
    """Specify the default CLUSTER_NAME which all operations should be run against.
    """
    cluster_resp = list_clusters()
    for cluster in cluster_resp.clusters:
        if cluster_name == cluster.name:
            cluster_id = cluster.id
            break
    else:
        raise click.ClickException(f"Cluster with name {cluster_name} not found")
    sdk_user.set_default_cluster(cluster_id)
    click.echo(f"✔ Successfully set the cluster context to {cluster_name}")



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1SessionStatus(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'destroyed_timestamp': 'datetime',
        'jupyter_lab_token': 'str',
        'jupyter_lab_url': 'str',
        'last_state_status_transition_timestamp': 'datetime',
        'message': 'str',
        'phase': 'V1SessionState',
        'reason': 'V1SessionStatusReason',
        'ssh_url': 'str',
        'start_timestamp': 'datetime',
        'stop_timestamp': 'datetime',
        'total_run_time_seconds': 'str'
    }

    attribute_map = {
        'destroyed_timestamp': 'destroyedTimestamp',
        'jupyter_lab_token': 'jupyterLabToken',
        'jupyter_lab_url': 'jupyterLabUrl',
        'last_state_status_transition_timestamp': 'lastStateStatusTransitionTimestamp',
        'message': 'message',
        'phase': 'phase',
        'reason': 'reason',
        'ssh_url': 'sshUrl',
        'start_timestamp': 'startTimestamp',
        'stop_timestamp': 'stopTimestamp',
        'total_run_time_seconds': 'totalRunTimeSeconds'
    }

    def __init__(self, destroyed_timestamp: 'datetime' = None, jupyter_lab_token: 'str' = None, jupyter_lab_url: 'str' = None, last_state_status_transition_timestamp: 'datetime' = None, message: 'str' = None, phase: 'V1SessionState' = None, reason: 'V1SessionStatusReason' = None, ssh_url: 'str' = None, start_timestamp: 'datetime' = None, stop_timestamp: 'datetime' = None, total_run_time_seconds: 'str' = None, _configuration=None):  # noqa: E501
        """V1SessionStatus - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._destroyed_timestamp = None
        self._jupyter_lab_token = None
        self._jupyter_lab_url = None
        self._last_state_status_transition_timestamp = None
        self._message = None
        self._phase = None
        self._reason = None
        self._ssh_url = None
        self._start_timestamp = None
        self._stop_timestamp = None
        self._total_run_time_seconds = None
        self.discriminator = None

        if destroyed_timestamp is not None:
            self.destroyed_timestamp = destroyed_timestamp
        if jupyter_lab_token is not None:
            self.jupyter_lab_token = jupyter_lab_token
        if jupyter_lab_url is not None:
            self.jupyter_lab_url = jupyter_lab_url
        if last_state_status_transition_timestamp is not None:
            self.last_state_status_transition_timestamp = last_state_status_transition_timestamp
        if message is not None:
            self.message = message
        if phase is not None:
            self.phase = phase
        if reason is not None:
            self.reason = reason
        if ssh_url is not None:
            self.ssh_url = ssh_url
        if start_timestamp is not None:
            self.start_timestamp = start_timestamp
        if stop_timestamp is not None:
            self.stop_timestamp = stop_timestamp
        if total_run_time_seconds is not None:
            self.total_run_time_seconds = total_run_time_seconds

    @property
    def destroyed_timestamp(self) -> 'datetime':
        """Gets the destroyed_timestamp of this V1SessionStatus.  # noqa: E501


        :return: The destroyed_timestamp of this V1SessionStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._destroyed_timestamp

    @destroyed_timestamp.setter
    def destroyed_timestamp(self, destroyed_timestamp: 'datetime'):
        """Sets the destroyed_timestamp of this V1SessionStatus.


        :param destroyed_timestamp: The destroyed_timestamp of this V1SessionStatus.  # noqa: E501
        :type: datetime
        """

        self._destroyed_timestamp = destroyed_timestamp

    @property
    def jupyter_lab_token(self) -> 'str':
        """Gets the jupyter_lab_token of this V1SessionStatus.  # noqa: E501


        :return: The jupyter_lab_token of this V1SessionStatus.  # noqa: E501
        :rtype: str
        """
        return self._jupyter_lab_token

    @jupyter_lab_token.setter
    def jupyter_lab_token(self, jupyter_lab_token: 'str'):
        """Sets the jupyter_lab_token of this V1SessionStatus.


        :param jupyter_lab_token: The jupyter_lab_token of this V1SessionStatus.  # noqa: E501
        :type: str
        """

        self._jupyter_lab_token = jupyter_lab_token

    @property
    def jupyter_lab_url(self) -> 'str':
        """Gets the jupyter_lab_url of this V1SessionStatus.  # noqa: E501


        :return: The jupyter_lab_url of this V1SessionStatus.  # noqa: E501
        :rtype: str
        """
        return self._jupyter_lab_url

    @jupyter_lab_url.setter
    def jupyter_lab_url(self, jupyter_lab_url: 'str'):
        """Sets the jupyter_lab_url of this V1SessionStatus.


        :param jupyter_lab_url: The jupyter_lab_url of this V1SessionStatus.  # noqa: E501
        :type: str
        """

        self._jupyter_lab_url = jupyter_lab_url

    @property
    def last_state_status_transition_timestamp(self) -> 'datetime':
        """Gets the last_state_status_transition_timestamp of this V1SessionStatus.  # noqa: E501


        :return: The last_state_status_transition_timestamp of this V1SessionStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._last_state_status_transition_timestamp

    @last_state_status_transition_timestamp.setter
    def last_state_status_transition_timestamp(self, last_state_status_transition_timestamp: 'datetime'):
        """Sets the last_state_status_transition_timestamp of this V1SessionStatus.


        :param last_state_status_transition_timestamp: The last_state_status_transition_timestamp of this V1SessionStatus.  # noqa: E501
        :type: datetime
        """

        self._last_state_status_transition_timestamp = last_state_status_transition_timestamp

    @property
    def message(self) -> 'str':
        """Gets the message of this V1SessionStatus.  # noqa: E501


        :return: The message of this V1SessionStatus.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message: 'str'):
        """Sets the message of this V1SessionStatus.


        :param message: The message of this V1SessionStatus.  # noqa: E501
        :type: str
        """

        self._message = message

    @property
    def phase(self) -> 'V1SessionState':
        """Gets the phase of this V1SessionStatus.  # noqa: E501


        :return: The phase of this V1SessionStatus.  # noqa: E501
        :rtype: V1SessionState
        """
        return self._phase

    @phase.setter
    def phase(self, phase: 'V1SessionState'):
        """Sets the phase of this V1SessionStatus.


        :param phase: The phase of this V1SessionStatus.  # noqa: E501
        :type: V1SessionState
        """

        self._phase = phase

    @property
    def reason(self) -> 'V1SessionStatusReason':
        """Gets the reason of this V1SessionStatus.  # noqa: E501

        Additional information about why the session is in some phase.  # noqa: E501

        :return: The reason of this V1SessionStatus.  # noqa: E501
        :rtype: V1SessionStatusReason
        """
        return self._reason

    @reason.setter
    def reason(self, reason: 'V1SessionStatusReason'):
        """Sets the reason of this V1SessionStatus.

        Additional information about why the session is in some phase.  # noqa: E501

        :param reason: The reason of this V1SessionStatus.  # noqa: E501
        :type: V1SessionStatusReason
        """

        self._reason = reason

    @property
    def ssh_url(self) -> 'str':
        """Gets the ssh_url of this V1SessionStatus.  # noqa: E501


        :return: The ssh_url of this V1SessionStatus.  # noqa: E501
        :rtype: str
        """
        return self._ssh_url

    @ssh_url.setter
    def ssh_url(self, ssh_url: 'str'):
        """Sets the ssh_url of this V1SessionStatus.


        :param ssh_url: The ssh_url of this V1SessionStatus.  # noqa: E501
        :type: str
        """

        self._ssh_url = ssh_url

    @property
    def start_timestamp(self) -> 'datetime':
        """Gets the start_timestamp of this V1SessionStatus.  # noqa: E501


        :return: The start_timestamp of this V1SessionStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._start_timestamp

    @start_timestamp.setter
    def start_timestamp(self, start_timestamp: 'datetime'):
        """Sets the start_timestamp of this V1SessionStatus.


        :param start_timestamp: The start_timestamp of this V1SessionStatus.  # noqa: E501
        :type: datetime
        """

        self._start_timestamp = start_timestamp

    @property
    def stop_timestamp(self) -> 'datetime':
        """Gets the stop_timestamp of this V1SessionStatus.  # noqa: E501


        :return: The stop_timestamp of this V1SessionStatus.  # noqa: E501
        :rtype: datetime
        """
        return self._stop_timestamp

    @stop_timestamp.setter
    def stop_timestamp(self, stop_timestamp: 'datetime'):
        """Sets the stop_timestamp of this V1SessionStatus.


        :param stop_timestamp: The stop_timestamp of this V1SessionStatus.  # noqa: E501
        :type: datetime
        """

        self._stop_timestamp = stop_timestamp

    @property
    def total_run_time_seconds(self) -> 'str':
        """Gets the total_run_time_seconds of this V1SessionStatus.  # noqa: E501


        :return: The total_run_time_seconds of this V1SessionStatus.  # noqa: E501
        :rtype: str
        """
        return self._total_run_time_seconds

    @total_run_time_seconds.setter
    def total_run_time_seconds(self, total_run_time_seconds: 'str'):
        """Sets the total_run_time_seconds of this V1SessionStatus.


        :param total_run_time_seconds: The total_run_time_seconds of this V1SessionStatus.  # noqa: E501
        :type: str
        """

        self._total_run_time_seconds = total_run_time_seconds

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1SessionStatus, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1SessionStatus') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1SessionStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1SessionStatus') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1SessionStatus):
            return True

        return self.to_dict() != other.to_dict()



import os
from pathlib import Path
import time
from typing import Optional

import click
from yaspin import yaspin

import grid.sdk.env as env
from grid.sdk.source_code.dependency_manager import (
    CONDA_MANAGER,
    CondaManager,
    DependencyManagerBase,
    find_package_manager,
    PIP_MANAGER,
    PipManager,
)
from grid.sdk.source_code.git import execute_git_command

PIP_REQUIREMENTS_FILE = "requirements.txt"
CONDA_REQUIREMENTS_FILE = "environment.yml"
WARNING_STR = click.style('WARNING', fg='yellow')


def wait_spinner(seconds=5):
    with yaspin():
        while seconds:
            time.sleep(1)
            seconds -= 1


class DependencyMixin:
    @staticmethod
    def _get_manager(repo_root) -> DependencyManagerBase:
        files = os.listdir(repo_root)
        conda_env = os.getenv("CONDA_DEFAULT_ENV")
        warning_str = click.style('WARNING', fg='yellow')
        if CONDA_REQUIREMENTS_FILE in files:
            if PIP_REQUIREMENTS_FILE in files:  # skipcq: PYL-R1705
                click.echo(
                    f"{warning_str} Found both {CONDA_REQUIREMENTS_FILE} "
                    f"and {PIP_REQUIREMENTS_FILE}. Defaulting to "
                    f"{PIP_REQUIREMENTS_FILE}"
                )
                manager_class = PipManager
            else:
                manager_class = CondaManager
        elif PIP_REQUIREMENTS_FILE in files:
            manager_class = PipManager
        elif conda_env:
            manager_class = CondaManager
        else:
            manager_class = PipManager
        return manager_class()

    @staticmethod
    def _guess_manager_from_dependency_file(dependency_file) -> DependencyManagerBase:
        """
        If user provides the dependency file instead of letting us find it from the root,
        use this function to create the Manager. If dependency_file is not provided,
        use `_get_manager`
        """
        package_manager = find_package_manager(dependency_file)
        if package_manager == PIP_MANAGER:
            return PipManager(dependency_file=dependency_file)
        if package_manager == CONDA_MANAGER:
            return CondaManager(dependency_file=dependency_file)

        raise RuntimeError("Unknown package manager")

    def serialize_dependencies(self,
                               config_path: Optional[str] = None,
                               dependency_file: Optional[str] = None) -> Optional[Path]:
        """
        Serialize dependency information to the corresponding source.

        Write python dependencies to the corresponding dependency listing
        (requirements.txt or environment.yml) and system dependencies to
        grid config (config.yml). Both write operations are actually offloaded
        downstream to the dependency manager object. Also fetches warnings
        from the dependency manager and echo to the user's terminal

        config:
            Grid config either read from the file provided by the user or the
            default generated one

        Returns
        -------
        Path to the requirement file
        """
        check_status = self._check_dependency_listing(do_diff_check=False, ignore_warnings=False)
        if not check_status:
            click.echo("Dependency check failed. Cannot sync environment")
            return None

        repo_root = execute_git_command(['rev-parse', '--show-toplevel'])
        try:
            if dependency_file:
                deps_manager = self._guess_manager_from_dependency_file(dependency_file)
            else:
                deps_manager = self._get_manager(repo_root)
        except Exception as e:  # noqa
            # it's already being checked by `check_dependency_listing`
            raise RuntimeError("Exception while initializing dependency manager")

        deps_manager.write_spec()
        deps_manager.write_config(config_path)
        return deps_manager.req_file

    def _check_dependency_listing(
        self,
        do_diff_check=True,
        ignore_warnings=env.IGNORE_WARNINGS,
        repo_root=execute_git_command(['rev-parse', '--show-toplevel']),
        dependency_file=None,
    ):
        """
        Evaluate if dependency listing is correct or not

        Determines the package manager employed by the user to specify
        application dependencies, setting up the appropriate interaction
        classes used to query one of the various package managers available
        as the system resolves dependency versions. Should no requirement file
        exist which specifies needed libraries+versions, this method provides
        the logic through which users are asked to select how they would like
        to proceed - automatic heuristic based generation or manual.

        Parameters
        ----------
        do_diff_check: bool
            Check for the diff between requirements file and the environment. This
            is True by default but `serialize_dependencies` checks this explicitly
            as part of serializing the dependencies and hence will set as False
        ignore_warnings: bool
            Should warnings be ignored. Should be removed when we move
            click component out of SDK -> TODO
        repo_root: str
            Root of the repository. Defaults to git root

        Returns
        -------
        Return False if check fails. True otherwise
        """
        if ignore_warnings:
            return True

        # Assuming the requirements file is always at the repo root if not passed
        # through dependency_file argument
        files = os.listdir(repo_root)

        try:
            if dependency_file:
                deps_manager = self._guess_manager_from_dependency_file(dependency_file)
            else:
                deps_manager = self._get_manager(repo_root)
        except Exception as e:  # noqa

            # ==============================================
            # Case 1: Initializing dependency manager failed
            # ==============================================
            # Ignoring to avoid internal errors to go to users
            if PIP_REQUIREMENTS_FILE not in files and CONDA_REQUIREMENTS_FILE not in files:
                message = f"""

        {WARNING_STR}
        No requirements.txt or environment.yml found. Your build could crash
        or not start.
        """

                click.echo(message)
                wait_spinner(seconds=2)
            return False

        # =====================================
        # Case 2: Requirements file not present
        # =====================================
        if PIP_REQUIREMENTS_FILE not in files and CONDA_REQUIREMENTS_FILE not in files:
            deps_found = "\n        ".join(deps_manager.source_deps)
            message = f"""

        {WARNING_STR}
        No requirements.txt or environment.yml found but we identified below
        dependencies from your source. Your build could crash or not
        start.

        {deps_found}
        """

            click.echo(message)
            wait_spinner(seconds=2)
            return False

        # ===========================================================
        # Case 3: Packages found imported for which spec is not found
        # ===========================================================
        missing = "\n        ".join(deps_manager.get_missing())
        if missing:
            message = f"""

        {WARNING_STR}

        Your requirements.txt or environment.yml are missing the following packages:

        {missing}

        Your experiment could crash if these are missing!
        Add them yourself or try `grid sync-env`
        """
            click.echo(message)
            wait_spinner(seconds=2)
            return False

        # ==============================================
        # Case 4: Incomplete/different-spec requirements
        # ==============================================
        if do_diff_check and deps_manager.has_change:
            message = f"""
        {WARNING_STR}
        Incomplete requirements.txt or environment.yml found
        Your build could crash or not start.

        Run `grid sync-env` to auto-populate the changed packages
        """
            click.echo(message)
            wait_spinner(seconds=2)
            return False

        # Returning True if all looks good
        return True



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1KubernetesClusterStatus(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'azure': 'V1AzureClusterDriverStatus',
        'bastion_host_ip': 'str',
        'eks': 'V1EKSCusterDriverStatus',
        'ingress_ca_cert': 'str',
        'ingress_load_balancer_hostname': 'str',
        'ingress_load_balancer_ip': 'str',
        'monitoring_client_cert': 'str',
        'monitoring_client_key': 'str',
        'ssh_private_key': 'str',
        'ssh_public_key': 'str'
    }

    attribute_map = {
        'azure': 'azure',
        'bastion_host_ip': 'bastionHostIp',
        'eks': 'eks',
        'ingress_ca_cert': 'ingressCaCert',
        'ingress_load_balancer_hostname': 'ingressLoadBalancerHostname',
        'ingress_load_balancer_ip': 'ingressLoadBalancerIp',
        'monitoring_client_cert': 'monitoringClientCert',
        'monitoring_client_key': 'monitoringClientKey',
        'ssh_private_key': 'sshPrivateKey',
        'ssh_public_key': 'sshPublicKey'
    }

    def __init__(self, azure: 'V1AzureClusterDriverStatus' = None, bastion_host_ip: 'str' = None, eks: 'V1EKSCusterDriverStatus' = None, ingress_ca_cert: 'str' = None, ingress_load_balancer_hostname: 'str' = None, ingress_load_balancer_ip: 'str' = None, monitoring_client_cert: 'str' = None, monitoring_client_key: 'str' = None, ssh_private_key: 'str' = None, ssh_public_key: 'str' = None, _configuration=None):  # noqa: E501
        """V1KubernetesClusterStatus - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._azure = None
        self._bastion_host_ip = None
        self._eks = None
        self._ingress_ca_cert = None
        self._ingress_load_balancer_hostname = None
        self._ingress_load_balancer_ip = None
        self._monitoring_client_cert = None
        self._monitoring_client_key = None
        self._ssh_private_key = None
        self._ssh_public_key = None
        self.discriminator = None

        if azure is not None:
            self.azure = azure
        if bastion_host_ip is not None:
            self.bastion_host_ip = bastion_host_ip
        if eks is not None:
            self.eks = eks
        if ingress_ca_cert is not None:
            self.ingress_ca_cert = ingress_ca_cert
        if ingress_load_balancer_hostname is not None:
            self.ingress_load_balancer_hostname = ingress_load_balancer_hostname
        if ingress_load_balancer_ip is not None:
            self.ingress_load_balancer_ip = ingress_load_balancer_ip
        if monitoring_client_cert is not None:
            self.monitoring_client_cert = monitoring_client_cert
        if monitoring_client_key is not None:
            self.monitoring_client_key = monitoring_client_key
        if ssh_private_key is not None:
            self.ssh_private_key = ssh_private_key
        if ssh_public_key is not None:
            self.ssh_public_key = ssh_public_key

    @property
    def azure(self) -> 'V1AzureClusterDriverStatus':
        """Gets the azure of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The azure of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: V1AzureClusterDriverStatus
        """
        return self._azure

    @azure.setter
    def azure(self, azure: 'V1AzureClusterDriverStatus'):
        """Sets the azure of this V1KubernetesClusterStatus.


        :param azure: The azure of this V1KubernetesClusterStatus.  # noqa: E501
        :type: V1AzureClusterDriverStatus
        """

        self._azure = azure

    @property
    def bastion_host_ip(self) -> 'str':
        """Gets the bastion_host_ip of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The bastion_host_ip of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._bastion_host_ip

    @bastion_host_ip.setter
    def bastion_host_ip(self, bastion_host_ip: 'str'):
        """Sets the bastion_host_ip of this V1KubernetesClusterStatus.


        :param bastion_host_ip: The bastion_host_ip of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """

        self._bastion_host_ip = bastion_host_ip

    @property
    def eks(self) -> 'V1EKSCusterDriverStatus':
        """Gets the eks of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The eks of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: V1EKSCusterDriverStatus
        """
        return self._eks

    @eks.setter
    def eks(self, eks: 'V1EKSCusterDriverStatus'):
        """Sets the eks of this V1KubernetesClusterStatus.


        :param eks: The eks of this V1KubernetesClusterStatus.  # noqa: E501
        :type: V1EKSCusterDriverStatus
        """

        self._eks = eks

    @property
    def ingress_ca_cert(self) -> 'str':
        """Gets the ingress_ca_cert of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The ingress_ca_cert of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._ingress_ca_cert

    @ingress_ca_cert.setter
    def ingress_ca_cert(self, ingress_ca_cert: 'str'):
        """Sets the ingress_ca_cert of this V1KubernetesClusterStatus.


        :param ingress_ca_cert: The ingress_ca_cert of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """
        if (self._configuration.client_side_validation and
                ingress_ca_cert is not None and not re.search(r'^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$', ingress_ca_cert)):  # noqa: E501
            raise ValueError(r"Invalid value for `ingress_ca_cert`, must be a follow pattern or equal to `/^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$/`")  # noqa: E501

        self._ingress_ca_cert = ingress_ca_cert

    @property
    def ingress_load_balancer_hostname(self) -> 'str':
        """Gets the ingress_load_balancer_hostname of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The ingress_load_balancer_hostname of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._ingress_load_balancer_hostname

    @ingress_load_balancer_hostname.setter
    def ingress_load_balancer_hostname(self, ingress_load_balancer_hostname: 'str'):
        """Sets the ingress_load_balancer_hostname of this V1KubernetesClusterStatus.


        :param ingress_load_balancer_hostname: The ingress_load_balancer_hostname of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """

        self._ingress_load_balancer_hostname = ingress_load_balancer_hostname

    @property
    def ingress_load_balancer_ip(self) -> 'str':
        """Gets the ingress_load_balancer_ip of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The ingress_load_balancer_ip of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._ingress_load_balancer_ip

    @ingress_load_balancer_ip.setter
    def ingress_load_balancer_ip(self, ingress_load_balancer_ip: 'str'):
        """Sets the ingress_load_balancer_ip of this V1KubernetesClusterStatus.


        :param ingress_load_balancer_ip: The ingress_load_balancer_ip of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """

        self._ingress_load_balancer_ip = ingress_load_balancer_ip

    @property
    def monitoring_client_cert(self) -> 'str':
        """Gets the monitoring_client_cert of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The monitoring_client_cert of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._monitoring_client_cert

    @monitoring_client_cert.setter
    def monitoring_client_cert(self, monitoring_client_cert: 'str'):
        """Sets the monitoring_client_cert of this V1KubernetesClusterStatus.


        :param monitoring_client_cert: The monitoring_client_cert of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """
        if (self._configuration.client_side_validation and
                monitoring_client_cert is not None and not re.search(r'^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$', monitoring_client_cert)):  # noqa: E501
            raise ValueError(r"Invalid value for `monitoring_client_cert`, must be a follow pattern or equal to `/^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$/`")  # noqa: E501

        self._monitoring_client_cert = monitoring_client_cert

    @property
    def monitoring_client_key(self) -> 'str':
        """Gets the monitoring_client_key of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The monitoring_client_key of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._monitoring_client_key

    @monitoring_client_key.setter
    def monitoring_client_key(self, monitoring_client_key: 'str'):
        """Sets the monitoring_client_key of this V1KubernetesClusterStatus.


        :param monitoring_client_key: The monitoring_client_key of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """
        if (self._configuration.client_side_validation and
                monitoring_client_key is not None and not re.search(r'^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$', monitoring_client_key)):  # noqa: E501
            raise ValueError(r"Invalid value for `monitoring_client_key`, must be a follow pattern or equal to `/^(?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{2}==|[A-Za-z0-9+\/]{3}=)?$/`")  # noqa: E501

        self._monitoring_client_key = monitoring_client_key

    @property
    def ssh_private_key(self) -> 'str':
        """Gets the ssh_private_key of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The ssh_private_key of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._ssh_private_key

    @ssh_private_key.setter
    def ssh_private_key(self, ssh_private_key: 'str'):
        """Sets the ssh_private_key of this V1KubernetesClusterStatus.


        :param ssh_private_key: The ssh_private_key of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """

        self._ssh_private_key = ssh_private_key

    @property
    def ssh_public_key(self) -> 'str':
        """Gets the ssh_public_key of this V1KubernetesClusterStatus.  # noqa: E501


        :return: The ssh_public_key of this V1KubernetesClusterStatus.  # noqa: E501
        :rtype: str
        """
        return self._ssh_public_key

    @ssh_public_key.setter
    def ssh_public_key(self, ssh_public_key: 'str'):
        """Sets the ssh_public_key of this V1KubernetesClusterStatus.


        :param ssh_public_key: The ssh_public_key of this V1KubernetesClusterStatus.  # noqa: E501
        :type: str
        """

        self._ssh_public_key = ssh_public_key

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1KubernetesClusterStatus, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1KubernetesClusterStatus') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1KubernetesClusterStatus):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1KubernetesClusterStatus') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1KubernetesClusterStatus):
            return True

        return self.to_dict() != other.to_dict()



from typing import Optional

import click

import grid.sdk.env as env
from grid.sdk.source_code.git import execute_git_command


class WorkflowChecksMixin:
    """Encapsulates workflow checks that prevent users from making mistakes."""
    @staticmethod
    def _check_if_remote_head_is_different() -> Optional[bool]:
        """
        Checks if remote git repository is different than
        the version available locally. This only compares the
        local SHA to the HEAD commit of a given branch. This
        check won't be used if user isn't in a HEAD locally.

        Original solution:

            * https://stackoverflow.com/questions/\
                3258243/check-if-pull-needed-in-git
        """
        # Check SHA values.
        local_sha = execute_git_command(['rev-parse', '@'])
        remote_sha = execute_git_command(['rev-parse', r"@{u}"])
        base_sha = execute_git_command(['merge-base', '@', r"@{u}"])

        # Whenever a SHA is not avaialble, just return.
        is_different = None
        if any('fatal' in f for f in (local_sha, remote_sha, base_sha)):
            return

        is_different = True
        if local_sha in (remote_sha, base_sha):
            is_different = False

        # Prompt user about the Github differences.
        warning_str = click.style('WARNING', fg='yellow')
        if is_different:
            message = f"{warning_str} Local git repository seems to be out-of-sync with GitHub. Continue?"
            if not env.IGNORE_WARNINGS:
                click.confirm(message, abort=True)

        return is_different

    @staticmethod
    def _check_github_repository() -> None:
        """Checks if the active directory is a GitHub repository."""
        github_repository = execute_git_command(["config", "--get", "remote.origin.url"])
        env.logger.debug(github_repository)

        if not github_repository or 'github.com' not in github_repository:
            # TODO - change error message to advertise `--localdir`
            raise click.ClickException(
                '`grid train` or `grid interactive` can only be run in a git repository '
                'hosted on github.com. See docs for details: https://docs.grid.ai'
            )

    @staticmethod
    def _check_if_uncommited_files() -> bool:
        """
        Checks if user has uncommited files in local repository.
        If there are uncommited files, then show a prompt
        indicating that uncommited files exist locally.

        Original solution:

            * https://stackoverflow.com/questions/\
                3878624/how-do-i-programmatically-determine-if-there-are-uncommitted-changes
        """
        files = execute_git_command(['update-index', '--refresh'])

        _files = files.split('\n')
        _files = [click.style(f"  {f.replace(': needs update', '')}", fg='red') for f in _files]

        files_str = '\n    '.join(_files)
        docs_str = click.style('https://docs.grid.ai', fg='blue')
        warning_str = click.style('WARNING', fg='yellow')
        message = f"""

    {warning_str}

    The following files are uncommited. Changes made to these
    files will not be avalable to Grid when running an Experiment.

    {files_str}

    Would you like to continue?

    You can use the flag --ignore_warnings to skip this warning.
    See details at: {docs_str}

   """
        has_uncommitted_files = False
        if files and not env.IGNORE_WARNINGS:
            click.confirm(message, abort=True)
            has_uncommitted_files = True

        return has_uncommitted_files



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

import requests
from tqdm import tqdm


@dataclass
class DownloadableObject:
    """Object that can be downloaded from an URL into a local path."""
    url: str
    download_path: str
    filename: str
    downloaded: bool = False


class Downloader:
    @staticmethod
    def single_file_download(
        obj: DownloadableObject,
        destination: Union[str, Path],
        progress_bar: Optional[tqdm] = None,
        chunk_size: int = 4096,  # golang bufio defaultBufSize
    ):
        """Downloads a single object to the filesystem.

        This function will produce one progress bar:
            1. Download progress of the file in bytes

        Parameters
        ----------
        obj : DownloadableObject
            The object to download.
        destination : Union[str, Path]
            Path in the filesystem to place download.
        progress_bar : Optional[tqdm], optional
            Existing progress bar, by default None.
            If None, a progress bar will be created.
            Shows bytes downloaded.
        chunk_size : int, optional
            We stream the file to disk in this chunk size (in bytes).
            By default 4096 bytes (4 kibibytes).
        """
        # make the destination directory if necessary
        destination = Path(destination) / Path(obj.download_path) / Path(obj.filename)
        Path(destination).parent.mkdir(parents=True, exist_ok=True)

        # request file to download
        file_download = requests.get(obj.url, allow_redirects=True, stream=True)
        file_download_size = int(file_download.headers.get("content-length", 0))

        # initialize progress bar
        if progress_bar is None:
            progress_bar = tqdm(unit="B", unit_scale=True)
        progress_bar.reset()
        progress_bar.set_description(desc=str(Path(obj.download_path) / obj.filename))
        progress_bar.total = file_download_size

        # start download
        with open(destination, 'wb') as file:
            for chunk in file_download.iter_content(chunk_size=chunk_size):
                if chunk:
                    progress_bar.update(file.write(chunk))
        progress_bar.refresh()

        # Mark object as downloaded
        obj.downloaded = True

    @staticmethod
    def multi_file_download(
        downloadable_objects: List[DownloadableObject],
        destination: str,
        description: str = "files downloaded",
        unit: str = "file",
        chunk_size: int = 4096,  # golang bufio defaultBufSize
        position: int = 0,
    ):
        """
        Downloads multiple files to the filesystem.

        This function will produce two progress bars:
            1. Progress of all files in `downloadable_objects`
            2. Download progress for a single file in bytes

        Attributes
        ----------
        downloadable_objects : List[DownloadableObject]
            List of objects containing URL and file system information.
        destination : str
            Path to place all downloads.
        description : str, optional
            String to display on the progress bar, by default "files downloaded".
        unt : str, optional
            Unit of the object we are downloading, by default "file".
            Some files have special names, like "artifact" or "log".
        chunk_size : int, optional
            We stream the file to disk in this chunk size (in bytes).
            By default 4096 bytes (4 kibibytes).
        position : int, optional
            The tqdm position argument, by default 0.
            Sets the order for multiple progress bars.
        """
        # It would be nice to keep the input as an iterator
        # but it could make the progress bars ugly.
        # You can't take the len() of a generator!
        downloadable_objects = list(downloadable_objects)

        total_progress = tqdm(downloadable_objects, unit=unit, position=position, desc=description, leave=False)
        file_progress = tqdm(unit="B", unit_scale=True, position=position + 1, leave=False)
        for obj in total_progress:
            obj: DownloadableObject

            # Download file to path
            # TODO figure out why concurrency was taken out in a previous commit
            Downloader.single_file_download(obj, destination, file_progress, chunk_size)

        file_progress.close()
        total_progress.close()



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1ComputeUsage(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'billed_time_seconds': 'str',
        'cluster_id': 'str',
        'cost': 'float',
        'created_at': 'datetime',
        'deleted_at': 'datetime',
        'hourly_cost': 'float',
        'id': 'str',
        'instance_type': 'str',
        'name': 'str',
        'resource_type': 'str',
        'resources': 'V1Resources',
        'user_id': 'str'
    }

    attribute_map = {
        'billed_time_seconds': 'billedTimeSeconds',
        'cluster_id': 'clusterId',
        'cost': 'cost',
        'created_at': 'createdAt',
        'deleted_at': 'deletedAt',
        'hourly_cost': 'hourlyCost',
        'id': 'id',
        'instance_type': 'instanceType',
        'name': 'name',
        'resource_type': 'resourceType',
        'resources': 'resources',
        'user_id': 'userId'
    }

    def __init__(self, billed_time_seconds: 'str' = None, cluster_id: 'str' = None, cost: 'float' = None, created_at: 'datetime' = None, deleted_at: 'datetime' = None, hourly_cost: 'float' = None, id: 'str' = None, instance_type: 'str' = None, name: 'str' = None, resource_type: 'str' = None, resources: 'V1Resources' = None, user_id: 'str' = None, _configuration=None):  # noqa: E501
        """V1ComputeUsage - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._billed_time_seconds = None
        self._cluster_id = None
        self._cost = None
        self._created_at = None
        self._deleted_at = None
        self._hourly_cost = None
        self._id = None
        self._instance_type = None
        self._name = None
        self._resource_type = None
        self._resources = None
        self._user_id = None
        self.discriminator = None

        if billed_time_seconds is not None:
            self.billed_time_seconds = billed_time_seconds
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if cost is not None:
            self.cost = cost
        if created_at is not None:
            self.created_at = created_at
        if deleted_at is not None:
            self.deleted_at = deleted_at
        if hourly_cost is not None:
            self.hourly_cost = hourly_cost
        if id is not None:
            self.id = id
        if instance_type is not None:
            self.instance_type = instance_type
        if name is not None:
            self.name = name
        if resource_type is not None:
            self.resource_type = resource_type
        if resources is not None:
            self.resources = resources
        if user_id is not None:
            self.user_id = user_id

    @property
    def billed_time_seconds(self) -> 'str':
        """Gets the billed_time_seconds of this V1ComputeUsage.  # noqa: E501


        :return: The billed_time_seconds of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._billed_time_seconds

    @billed_time_seconds.setter
    def billed_time_seconds(self, billed_time_seconds: 'str'):
        """Sets the billed_time_seconds of this V1ComputeUsage.


        :param billed_time_seconds: The billed_time_seconds of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._billed_time_seconds = billed_time_seconds

    @property
    def cluster_id(self) -> 'str':
        """Gets the cluster_id of this V1ComputeUsage.  # noqa: E501


        :return: The cluster_id of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id: 'str'):
        """Sets the cluster_id of this V1ComputeUsage.


        :param cluster_id: The cluster_id of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def cost(self) -> 'float':
        """Gets the cost of this V1ComputeUsage.  # noqa: E501


        :return: The cost of this V1ComputeUsage.  # noqa: E501
        :rtype: float
        """
        return self._cost

    @cost.setter
    def cost(self, cost: 'float'):
        """Sets the cost of this V1ComputeUsage.


        :param cost: The cost of this V1ComputeUsage.  # noqa: E501
        :type: float
        """

        self._cost = cost

    @property
    def created_at(self) -> 'datetime':
        """Gets the created_at of this V1ComputeUsage.  # noqa: E501


        :return: The created_at of this V1ComputeUsage.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at: 'datetime'):
        """Sets the created_at of this V1ComputeUsage.


        :param created_at: The created_at of this V1ComputeUsage.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def deleted_at(self) -> 'datetime':
        """Gets the deleted_at of this V1ComputeUsage.  # noqa: E501


        :return: The deleted_at of this V1ComputeUsage.  # noqa: E501
        :rtype: datetime
        """
        return self._deleted_at

    @deleted_at.setter
    def deleted_at(self, deleted_at: 'datetime'):
        """Sets the deleted_at of this V1ComputeUsage.


        :param deleted_at: The deleted_at of this V1ComputeUsage.  # noqa: E501
        :type: datetime
        """

        self._deleted_at = deleted_at

    @property
    def hourly_cost(self) -> 'float':
        """Gets the hourly_cost of this V1ComputeUsage.  # noqa: E501


        :return: The hourly_cost of this V1ComputeUsage.  # noqa: E501
        :rtype: float
        """
        return self._hourly_cost

    @hourly_cost.setter
    def hourly_cost(self, hourly_cost: 'float'):
        """Sets the hourly_cost of this V1ComputeUsage.


        :param hourly_cost: The hourly_cost of this V1ComputeUsage.  # noqa: E501
        :type: float
        """

        self._hourly_cost = hourly_cost

    @property
    def id(self) -> 'str':
        """Gets the id of this V1ComputeUsage.  # noqa: E501


        :return: The id of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id: 'str'):
        """Sets the id of this V1ComputeUsage.


        :param id: The id of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def instance_type(self) -> 'str':
        """Gets the instance_type of this V1ComputeUsage.  # noqa: E501


        :return: The instance_type of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type: 'str'):
        """Sets the instance_type of this V1ComputeUsage.


        :param instance_type: The instance_type of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._instance_type = instance_type

    @property
    def name(self) -> 'str':
        """Gets the name of this V1ComputeUsage.  # noqa: E501


        :return: The name of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: 'str'):
        """Sets the name of this V1ComputeUsage.


        :param name: The name of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def resource_type(self) -> 'str':
        """Gets the resource_type of this V1ComputeUsage.  # noqa: E501


        :return: The resource_type of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._resource_type

    @resource_type.setter
    def resource_type(self, resource_type: 'str'):
        """Sets the resource_type of this V1ComputeUsage.


        :param resource_type: The resource_type of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._resource_type = resource_type

    @property
    def resources(self) -> 'V1Resources':
        """Gets the resources of this V1ComputeUsage.  # noqa: E501


        :return: The resources of this V1ComputeUsage.  # noqa: E501
        :rtype: V1Resources
        """
        return self._resources

    @resources.setter
    def resources(self, resources: 'V1Resources'):
        """Sets the resources of this V1ComputeUsage.


        :param resources: The resources of this V1ComputeUsage.  # noqa: E501
        :type: V1Resources
        """

        self._resources = resources

    @property
    def user_id(self) -> 'str':
        """Gets the user_id of this V1ComputeUsage.  # noqa: E501


        :return: The user_id of this V1ComputeUsage.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id: 'str'):
        """Sets the user_id of this V1ComputeUsage.


        :param user_id: The user_id of this V1ComputeUsage.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1ComputeUsage, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1ComputeUsage') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ComputeUsage):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1ComputeUsage') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ComputeUsage):
            return True

        return self.to_dict() != other.to_dict()



import json

import click

from grid.cli import rich_click
from grid.cli.cli.utilities import create_openapi_object, string2dict
from grid.openapi import Body
from grid.openapi.api import cluster_service_api
from grid.sdk.client import create_swagger_client
from grid.cli.cli.grid_clusters import assert_instance_types_are_supported


@rich_click.group()
def edit() -> None:
    """Edits a resource"""
    pass


@edit.command()
@rich_click.argument('cluster', type=str)
def cluster(cluster: str):
    """Edit existing cluster"""
    api_client = create_swagger_client()
    cluster_api = cluster_service_api.ClusterServiceApi(api_client)
    resp = cluster_api.cluster_service_get_cluster(cluster)
    spec_str = click.edit(json.dumps(resp.spec.to_dict(), indent=4))
    new_spec = resp.spec
    if spec_str is not None:
        new_spec = create_openapi_object(string2dict(spec_str), resp.spec)
    if new_spec == resp.spec:
        click.echo("cluster unchanged")
        return

    # Check the instance types are of supported architecture
    assert_instance_types_are_supported([i.name for i in new_spec.driver.kubernetes.aws.instance_types])

    body = Body(name=cluster, spec=new_spec)
    update_resp = cluster_api.cluster_service_update_cluster(id=cluster, body=body)
    click.echo(update_resp.to_str())



from grid.cli.cli.grid_artifacts import artifacts
from grid.cli.cli.grid_clusters import clusters
from grid.cli.cli.grid_credentials import credential
from grid.cli.cli.grid_datastore import datastore
from grid.cli.cli.grid_delete import delete
from grid.cli.cli.grid_edit import edit
from grid.cli.cli.grid_env import sync_env
from grid.cli.cli.grid_history import history
from grid.cli.cli.grid_instance_type import instance_types
from grid.cli.cli.grid_login import login
from grid.cli.cli.grid_logs import logs
from grid.cli.cli.grid_run import run
from grid.cli.cli.grid_session import session
from grid.cli.cli.grid_ssh_keys import ssh_keys
from grid.cli.cli.grid_status import status
from grid.cli.cli.grid_stop import stop
from grid.cli.cli.grid_team import team
from grid.cli.cli.grid_user import user
from grid.cli.cli.grid_view import view

__all__ = [
    'view',
    'status',
    'login',
    'run',
    'stop',
    'credential',
    'history',
    'logs',
    'session',
    'artifacts',
    'delete',
    'datastore',
    'ssh_keys',
    'user',
    'sync_env',
    'clusters',
    'edit',
    'team',
    'instance_types',
]



import json
import logging
import os
from pathlib import Path


def write_default_settings(p: Path):
    p.parent.mkdir(parents=True, exist_ok=True)
    global_variables = {
        'debug': False,
        'ignore_warnings': False,
        'skip_vesion_check': False,
        'context': 'prod-2',  # TODO - we shouldn't default to an ID but to the name - `grid-cloud`
        'grid_url': 'https://platform.grid.ai'
    }
    with P.open('w') as file:
        json.dump(global_variables, file, ensure_ascii=False, indent=4)


def reset_global_variables() -> None:
    """ Reset the settings from env variables"""
    global DEBUG, SKIP_VERSION_CHECK, CONTEXT, IGNORE_WARNINGS, GRID_URL

    if 'DEBUG' in os.environ:
        DEBUG = bool(os.environ['DEBUG'])

    if 'IGNORE_WARNINGS' in os.environ:
        IGNORE_WARNINGS = bool(os.environ['IGNORE_WARNINGS'])

    if 'SKIP_VERSION_CHECK' in os.environ:
        SKIP_VERSION_CHECK = bool(os.environ['SKIP_VERSION_CHECK'])

    if 'GRID_CLUSTER_ID' in os.environ:
        CONTEXT = os.environ['GRID_CLUSTER_ID']

    if 'GRID_URL' in os.environ:
        GRID_URL = os.environ['GRID_URL']


P = Path.home().joinpath(".grid/settings.json")
# Make sure path exists.
Path(P.parents[0]).mkdir(parents=True, exist_ok=True)
# If file doesn't exist, create with default global settings
if not P.exists():
    write_default_settings(P)
user_settings = json.load(P.open())

# TODO ENV variables could lead into confusion
DEBUG = bool(os.getenv("DEBUG", user_settings.get('debug', False)))

IGNORE_WARNINGS = bool(os.getenv('IGNORE_WARNINGS', user_settings.get('ignore_warnings', False)))
SKIP_VERSION_CHECK = bool(os.getenv('SKIP_VERSION_CHECK', user_settings.get('skip_vesion_check', False)))

CONTEXT = os.getenv('GRID_CLUSTER_ID', user_settings.get('context', 'prod-2'))
GRID_DIR = os.getenv('GRID_DIR', str(Path.home() / ".grid"))
GRID_URL = os.getenv("GRID_URL", user_settings.get('grid_url', 'https://platform.grid.ai'))
GRID_CREDENTIAL_PATH = os.getenv('GRID_CREDENTIAL_PATH', str(Path.home() / '.grid' / 'credentials.json'))
GRID_SETTINGS_PATH = os.getenv('GRID_SETTINGS_PATH', str(Path.home() / '.grid' / 'settings.json'))

logger = logging.getLogger(__name__)  # pragma: no cover

SHOW_PROCESS_STATUS_DETAILS = os.getenv('SHOW_PROCESS_STATUS_DETAILS')
GRID_SKIP_GITHUB_TOKEN_CHECK = bool(os.getenv("GRID_SKIP_GITHUB_TOKEN_CHECK", default=None))
TESTING = os.getenv("TESTING", default=os.getenv("CI", default=None))
SSL_CA_CERT = os.getenv("REQUESTS_CA_BUNDLE", default=os.getenv("SSL_CERT_FILE", default=None))
GRID_SSH_CONFIG = os.getenv("GRID_SSH_CONFIG", default=str(Path.home() / ".ssh/config"))

_KB = 1024
_MB = _KB**2
_GB = _KB**3
_TB = _KB**4

MAX_BYTES_PER_FILE_PART_UPLOAD = int(os.getenv("GRID_DATASTORE_MAX_BYTES_PER_FILE_PART_UPLOAD", default=50 * _MB))
MAX_BYTES_PER_BATCH_UPLOAD = int(os.getenv("GRID_DATASTORE_MAX_BYTES_PER_BATCH_UPLOAD", default=2 * _GB))
MAX_FILES_PER_UPLOAD_BATCH = int(os.getenv("GRID_DATASTORE_MAX_FILES_PER_UPLOAD_BATCH", default=3000))
MAX_FILES_PER_FINALIZE_BATCH = int(os.getenv("GRID_DATASTORE_MAX_FILES_PER_FINALIZE_BATCH", default=3000))

UPLOAD_MAX_WORKER_THREADS = int(os.getenv("GRID_DATASTORE_UPLOAD_MAX_WORKER_THREADS", 20))



from collections import OrderedDict
from pathlib import Path
import shlex
import sys
from typing import Dict, Optional

import click
import yaspin

from grid import Run, Resources, ScratchSpace
from grid.cli import rich_click
from grid.cli.cli.utilities import verify_entrypoint_script
import grid.sdk.env as env
from grid.sdk.rest.exceptions import GridException
from grid.sdk.datastores import fetch_datastore
from grid.sdk.utils.name_generator import unique_name


def _check_run_name_is_valid(_ctx, _param, value):
    """Click callback that checks if a Run contains reserved names."""
    if value is not None:
        fail = False

        #  Check if the input is alphanumeric.
        _run_name = value.replace('-', '')
        if not _run_name.isalnum():
            fail = True

        #  Check if the allowed `-` character is not used
        #  at the end of the string.
        elif value.endswith('-') or value.startswith('-'):
            fail = True

        #  Check that the run name does not contain any
        #  uppercase characters.
        elif any(x.isupper() for x in value):
            fail = True

        if fail:
            raise click.BadParameter(
                f"""

            Invalid Run name: {value} the Run name must be lower case
            alphanumeric characters or '-', start with an alphabetic
            character, and end with an alphanumeric character (e.g. 'my-name',
            or 'abc-123').

                """
            )

    return value


def _check_ignore_warnings_flag(ctx, _param, value):
    """
    Click callback that assigns the value of the ignore warnings callback
    to a global variable.
    """
    if value is not None:
        env.IGNORE_WARNINGS = value
    return value


def _aws_node_to_nickname():
    aws_node_to_nicknames = OrderedDict({
        # AWS
        'p3.16xlarge': '8_v100_16gb',
        'p3dn.24xlarge': '8_v100_32gb',
        'g4dn.metal': '8_t4_16gb',
        'p2.8xlarge': '8_k80_12gb',
        'p3.8xlarge': '4_v100_16gb',
        'g4dn.12xlarge': '4_t4_16gb',
        'g3.16xlarge': '4_m60_8gb',
        'g3.8xlarge': '2_m60_8gb',
        'p3.2xlarge': '1_v100_16gb',
        # 'p4d.24xlarge': '8_a100_40gb',  # currently not supported
        'g4dn.8xlarge': '1_t4_16gb',
        'g4dn.4xlarge': '1_t4_16gb',
        'g4dn.2xlarge': '1_t4_16gb',
        'g4dn.xlarge': '1_t4_16gb',
        'g4dn.16xlarge': '1_t4_16gb',
        'p2.xlarge': '1_k80_12gb',
        'g3s.xlarge': '1_m60_8gb',
        'g3.4xlarge': '1_m60_8gb',
        'm5a.large': '2_cpu_8gb',
        't2.large': '2_cpu_8gb',
        't2.medium': '2_cpu_4gb'
    })
    return aws_node_to_nicknames


def _nickname_to_aws_nodes():
    aws_node_to_nickname = _aws_node_to_nickname()
    aws_nickname_to_node = {v: k for k, v in aws_node_to_nickname.items()}
    return aws_nickname_to_node


def _resolve_instance_type_nickname(ctx, _param, value):
    """
    Enables instance type shortcuts like:
    2_cpu_4gb for t2.large
    """
    nickname = value.lower()

    aws_nickname_to_node = _nickname_to_aws_nodes()

    # validate potential options for the node name
    possible_values = list(aws_nickname_to_node.keys()) + list(aws_nickname_to_node.values())
    if nickname not in possible_values:
        possible_options = '\n'.join(list(aws_nickname_to_node.keys()))
        click.BadParameter(f'{nickname} is not an available instance_type\n try one of these:\n{possible_options}')

    instance_type = nickname

    # if the value has _ then it's a nickname
    if '_' in nickname:
        instance_type = aws_nickname_to_node[nickname]
    return instance_type


def _get_instance_types(ctx, args, incomplete):
    # TODO: these should be retrieved from backend
    return list(_aws_node_to_nickname().keys())


def _run_command_builder(ctx, _param, value):
    # Captures invocation command exactly as it was
    # typed in the terminal. pipe.quote maintains original
    # quotation used in the CLI. We'll replace absolute
    # grid executable path with the `grid` alias.
    script = value[0]
    verify_entrypoint_script(script)
    sys.argv[0] = "grid"
    invocation_command = " ".join(map(shlex.quote, sys.argv))
    _, script_args = invocation_command.split(script)
    return f"{script} {script_args}"


@rich_click.command(context_settings=dict(ignore_unknown_options=True))
@click.option('--config', 'config', type=Path, required=False, help='Path to Grid config YML.')
@click.option('--name', 'name', type=str, required=False, help='Name for this run', callback=_check_run_name_is_valid)
@click.option('--cluster', 'cluster', type=str, default=env.CONTEXT, required=False)
@click.option(
    '--strategy',
    'strategy',
    type=click.Choice(['grid_search', 'random_search', 'none'], case_sensitive=False),
    required=False,
    default="grid_search",
    help="""Hyper-parameter search strategy\b

Available options:
 - grid_search (creates invocations for each combinations or evaluated arguments)
 - random_search (creates random_search conbination of parameters - you can add num_trials for defining exact number of them)
 - none (script arguments will not be evaluated at all and passed directly to the script)
"""
)
@click.option(
    '--num_trials',
    'num_trials',
    type=str,
    required=False,
    help='Number of samples from full search space that are used by the random_search strategy'
)
@click.option(
    '--seed',
    'seed',
    type=str,
    required=False,
    help='Seed value for the `random_search` strategy',
)
@click.option(
    '--instance_type',
    'instance_type',
    type=str,
    default='m5a.large',
    help='Instance type to start training session in',
    callback=_resolve_instance_type_nickname
)
@click.option('--gpus', 'gpus', type=int, required=False, default=0, help='Number of GPUs to allocate per experiment')
@click.option('--cpus', 'cpus', type=int, required=False, default=0, help='Number of CPUs to allocate per experiment')
@click.option('--memory', 'memory', type=str, default="100", required=False, help='How much memory an experiment needs')
@click.option(
    '--datastore_name', 'datastore_name', type=str, required=False, help='Datastore name to be mounted in training'
)
@click.option(
    '--datastore_version',
    'datastore_version',
    type=int,
    required=False,
    help='Datastore version to be mounted in training'
)
@click.option(
    '--datastore_mount_dir',
    'datastore_mount_dir',
    type=str,
    required=False,
    help='Directory to mount Datastore in training job. The default datastore mount location is /datastores'
)
@click.option(
    '--framework',
    'framework',
    required=False,
    default='lightning',
    help=
    "Framework to use in training. Select from available options: `lightning`, `torch`, `tensorflow`, `julia (will select the latest available version)`, `julia:1.6.1`, `julia:1.6.2`, `julia:1.6.3`, `julia:1.6.4`, `julia:1.6.5`, `julia:1.7.0`, `julia:1.7.1`, `torchelastic`",  # noqa
)
@click.option(
    '--use_spot',
    'use_spot',
    is_flag=True,
    required=False,
    help='Use spot instance. The spot instances, or preemptive instance can be shut down at will'
)
@click.option(
    '--ignore_warnings',
    is_flag=True,
    required=False,
    help='If we should ignore warning when executing commands',
    callback=_check_ignore_warnings_flag
)
@click.option(
    '--scratch_size',
    type=int,
    default=100,
    required=False,
    help='The size in GB of the scratch space attached to the experiment'
)
@click.option(
    '--scratch_mount_path',
    type=str,
    required=False,
    help='The mount path to mount the scratch space',
    default='/tmp/scratch',
)
@click.option(
    '-l',
    '--localdir',
    is_flag=True,
    required=False,
    help="""Upload source code from the local directory instead of having Grid clone the repo from GitHub (default).
This option is particularly useful for users that do not host their source code on GitHub.
"""
)
@click.option('-d', '--dockerfile', type=str, required=False, help="Dockerfile for the image building")
@click.option(
    '--dependency_file',
    type=str,
    required=False,
    help="""Dependency file path. \b

If not provided and a `requirements.txt`, `environment.yml`, or `Project.toml`
file is present in the current-working-directory, then we will automatically
install dependencies from according to the inferred file."""
)
@click.option(
    '--auto_resume',
    'auto_resume',
    is_flag=True,
    required=False,
    help="""Mark this run as auto-resumable. \b

If underlying node/instance/VM is terminated, the experiment will be
automatically resumed, with all artifacts restores from the last
known state. The experiment code will receive SIGTERM signal and it
must exit with status code 0 upon properly dumping its state to disk.
"""
)
@rich_click.argument(
    'run_command', nargs=-1, callback=_run_command_builder, help='Arguments to be passed to the script.'
)
def run(
    run_command: str,
    config: Optional[Dict],
    name: Optional[str],
    cluster: Optional[str],
    strategy: Optional[str],
    num_trials: Optional[int],
    seed: Optional[str],
    instance_type: str,
    gpus: Optional[int],
    ignore_warnings: bool,
    memory: Optional[str],
    cpus: Optional[int],
    datastore_name: Optional[str],
    datastore_version: Optional[int],
    datastore_mount_dir: Optional[str],
    framework: Optional[str],
    use_spot: Optional[bool],
    scratch_size: Optional[int],
    scratch_mount_path: Optional[str],
    localdir: Optional[bool],
    dockerfile: Optional[str],
    dependency_file: Optional[str],
    auto_resume: Optional[bool] = None,
) -> None:
    """Launch a Run from some SCRIPT with the provided SCRIPT_ARGS.

    A run is a collection of experiments which run with a single set of SCRIPT_ARGS. The
    SCRIPT_ARGS passed to the run command can represent fixed values, or a set of values
    to be searched over for each option. If a set of values are passed, a sweep (grid-search
    or random-search) will be performed, launching the desired number of experiments in
    parallel - each with a unique set of input arguments.

    The script runs on the specified instance type and Grid collects the generated
    artifacts, metrics, and logs; making them available for you to view in real time
    (or later if so desired) on either our Web UI or via this CLI.
    """

    # Setting this to the global variable so downstream tasks can pick it up easily
    if ignore_warnings:
        env.IGNORE_WARNINGS = ignore_warnings
    validate_xpus(instance_type=instance_type, xpu=gpus or cpus)

    if env.DEBUG:
        click.echo(f"Run command: {run_command}")
        click.echo(f"Hyperparams Search Strategy: {strategy}")
        click.echo(f"GPUs Requested: {gpus}")
        click.echo(f"Spot Requested: {use_spot}")

    # make a fun random name when user does not pass in a name
    if name is None:
        name = unique_name()

    dstore = None
    if datastore_name:
        dstore = fetch_datastore(datastore_name, datastore_version, cluster)

    # these values will be overwritten by values in the config file if they exist
    resources = Resources(instance_type=instance_type, gpus=gpus, cpus=cpus, storage_gb=int(memory), use_spot=use_spot)
    scratch = ScratchSpace(size_gb=scratch_size, mount_path=scratch_mount_path)

    # starting the spinner
    spinner = yaspin.yaspin(text=f"Submitting Run {name} ...", color="yellow")
    spinner.start()
    try:
        run_obj = Run(
            name=name,
            run_command=run_command,
            strategy=strategy,
            strategy_options={} if strategy != 'random_search' else {
                'num_trials': num_trials or '',
                'seed': seed or '0',
            },
            framework=framework,
            dependency_file=dependency_file,
            localdir=localdir,
            dockerfile=dockerfile,
            auto_resume=auto_resume,
            resources=resources,
            datastore=dstore,
            datastore_mount_dir=datastore_mount_dir,
            scratch=scratch,
            cluster_id=cluster,
            config_file=config,
        )

        # when starting a localdir run, the local directory is uploaded to s3.
        # The upload bar and the spinner interfere, causing lots of noise
        if localdir:
            spinner.stop()

        if run_obj._is_created:
            raise click.ClickException(
                f"A Run with the name {name} already exists. Please specify a unique name and try submitting your run again. "
            )

        run_obj.start()
    except GridException as e:
        spinner.fail("✘")
        raise click.ClickException(str(e))
    finally:
        # stopping the spinner
        spinner.stop()

    if run_obj.exists:
        datastore_name = run_obj.datastore.name if run_obj.datastore else "None"
        datastore_version = run_obj.datastore.version if run_obj.datastore else "None"
        datastore_mount_dir = run_obj.datastore_mount_dir if run_obj.datastore_mount_dir else "None"
        message = f"""
                    Run submitted!
                    `grid status` to list all runs
                    `grid status {run_obj.name}` to see all experiments for this run

                    ----------------------
                    Submission summary
                    ----------------------
                    name:                    {run_obj.name}
                    run_command:             {run_obj.run_command}
                    instance_type:           {run_obj.resources.instance_type}
                    use_spot:                {run_obj.resources.use_spot}
                    cluster:                 {run_obj.cluster_id}
                    datastore_name:          {datastore_name}
                    datastore_version:       {datastore_version}
                    datastore_mount_dir:     {datastore_mount_dir}
                    """
        click.echo(message)
    else:
        click.echo(f"Run {name} failed to submit")


def validate_xpus(instance_type: str, xpu: int) -> None:
    if not instance_type:
        return

    aws_machine_name = _aws_node_to_nickname().get(instance_type, None)
    if not aws_machine_name:
        return

    machine_xpus = aws_machine_name.split('_')[0]
    assert machine_xpus.isdigit(), "aws machine names formatting is not correct, it should be <xpu-count>_name_<memory>"

    if xpu and int(machine_xpus) < xpu:
        raise click.exceptions.ClickException(
            f"You have specified {xpu} GPUs (or CPUs) which is more than available on your selected instance type.\nPlease specify up to {int(machine_xpus)} GPUs and try again."
        )



from datetime import datetime, timezone
from typing import Dict, Optional

import click
from gql import Client, gql

from grid import list_runs
from grid.cli.observables.base import BaseObservable
from grid.cli.utilities import get_abs_time_difference, string_format_timedelta
import grid.sdk.env as env
from grid.sdk.runs import RunState


class Run(BaseObservable):
    def __init__(self, identifier: Optional[str] = []):
        self.identifier = identifier
        super().__init__(spinner_load_type="Runs")

    def get(self, is_global: Optional[bool] = False):
        """
        Gets the run status; either for a single run or all runs for
        user.

        Parameters
        ----------
        is_global: Optional[bool], default False
            if True, returns status of experiment(s) of the everyone in the team (TODO)
        run_name:
            name of the run to get status of
        """
        self.spinner.start()
        self.spinner.text = 'Getting Run status...'
        try:
            # TODO - show runs from all the clusters
            runs = list_runs(is_global=is_global)
            self.spinner.text = 'Done!'
        except Exception as e:
            self.spinner.fail("✘")
            self.spinner.stop()
            if env.DEBUG:
                click.echo(str(e))
            raise click.ClickException(f'Query to Grid failed. {e}')

        if is_global:
            table_cols = [
                'Run',
                'Status',
                'Created By',
                'Duration',
                'Experiments',
                'Running',
                'Queued',
                'Completed',
                'Failed',
                'Stopped',
            ]
        else:
            table_cols = [
                'Run',
                'Status',
                'Duration',
                'Experiments',
                'Running',
                'Queued',
                'Completed',
                'Failed',
                'Stopped',
            ]

        table = BaseObservable.create_table(columns=table_cols)

        #  Whenever we don't have yet submitted experiments,
        table_rows = 0
        for run in runs:
            # we only have 3 statuses for runs
            # running (if something is running)
            # TODO - move status to Enum
            is_running = run.status == "RUN_STATE_RUNNING"

            # We classify queued and pending into queued
            # TODO - standardize the string here and the actual status
            n_queued = run.experiment_counts.get("pending", 0)
            n_running = run.experiment_counts.get("running", 0)
            n_failed = run.experiment_counts.get("failed", 0)
            n_succeeded = run.experiment_counts.get("succeeded", 0)
            n_cancelled = run.experiment_counts.get("cancelled", 0)

            # If anything is queued, the the status of the entire
            # run is queued. All other statuses are running in
            # all other conditions.
            if n_queued > 0:
                status = 'queued'

            # If you have anything running (and nothing queued)
            # then, mark the run as running.
            elif is_running:
                status = 'running'

            # If it doesn't match the conditions above, just
            # skip this row and add the row and put it in history.
            else:
                # don't render table because it should be in history
                continue

            # Calculate the duration column
            delta = get_abs_time_difference(datetime.now(timezone.utc), run.created_at)
            duration_str = string_format_timedelta(delta)

            if is_global:
                table.add_row(
                    run.name, status, run.user.username, duration_str, str(len(run.experiments)), str(), str(n_queued),
                    str(n_succeeded), str(n_failed), str(n_cancelled)
                )
            else:
                table.add_row(
                    run.name, status, duration_str, str(len(run.experiments)), str(n_running), str(n_queued),
                    str(n_succeeded), str(n_failed), str(n_cancelled)
                )

            #  Let's count how many rows have been added.
            table_rows += 1

        #  Close the spinner.
        self.spinner.ok("✔")
        self.spinner.stop()

        # If there are no Runs to render, add a
        # placeholder row indicating none are active.
        if table_rows == 0:
            table.add_row("None Active.", *[" " for i in range(len(table_cols) - 1)])

        self.console.print(table)

        #  Print useful message indicating that users can run
        #  grid history.
        history_runs = len(runs) - table_rows
        if history_runs > 0:
            click.echo(f'{history_runs} Run(s) are not active. Use `grid history` ' 'to view your Run history.\n')

        return {'getRuns': runs}

    def get_history(self, is_global: Optional[bool] = False, run_name: Optional[str] = None):
        """
        Fetches a complete history of runs. This includes runs that
        are not currently active.

        Parameters
        ----------
        is_global: Optional[bool], default False
            Returns status of session from everyone in the team
        """
        self.spinner.start()
        self.spinner.text = 'Getting Runs ...'
        try:
            # TODO - show runs from all the clusters
            runs = list_runs(is_global=is_global)
            self.spinner.text = 'Done!'
        except Exception as e:
            self.spinner.fail("✘")
            self.spinner.stop()
            if env.DEBUG:
                click.echo(str(e))
            raise click.ClickException(f'Query to Grid failed. {e}')

        if is_global:
            table_cols = ['Run', 'Created By', 'Created At', 'Experiments', 'Failed', 'Stopped', 'Completed']
        else:
            table_cols = ['Run', 'Created At', 'Experiments', 'Failed', 'Stopped', 'Completed']
        table = BaseObservable.create_table(columns=table_cols)

        for run in runs:
            if run_name and run_name != run.name:
                continue
            n_failed = run.experiment_counts.get(RunState.FAILED.value, 0)
            n_succeeded = run.experiment_counts.get(RunState.SUCCEEDED.value, 0)
            n_cancelled = run.experiment_counts.get(RunState.CANCELLED.value, 0)

            # Calculate the duration column
            delta = get_abs_time_difference(datetime.now(timezone.utc), run.created_at)
            duration_str = string_format_timedelta(delta)

            if is_global:
                table.add_row(
                    run.name, run.user.username, duration_str, str(len(run.experiments)), str(n_failed),
                    str(n_cancelled), str(n_succeeded)
                )
            else:
                table.add_row(
                    run.name, duration_str, str(len(run.experiments)), str(n_failed), str(n_cancelled),
                    str(n_succeeded)
                )

        #  Close the spinner.
        self.spinner.ok("✔")
        self.spinner.stop()

        # Add placeholder row if no records are available.
        if not runs:
            table.add_row("No History.", *[" " for i in range(len(table_cols) - 1)])

        self.console.print(table)

        return runs

    def follow(self, is_global: Optional[bool] = False):  # pragma: no cover
        pass
